import asyncio
import time
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.derivative.backpack_perpetual import (
    backpack_perpetual_constants as CONSTANTS,
    backpack_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.backpack_perpetual.backpack_perpetual_order_book import BackpackPerpetualOrderBook
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.backpack_perpetual.backpack_perpetual_derivative import (
        BackpackPerpetualDerivative,
    )


class BackpackPerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'BackpackPerpetualDerivative',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__(trading_pairs)
        self._connector = connector
        self._trade_messages_queue_key = CONSTANTS.TRADE_EVENT_TYPE
        self._diff_messages_queue_key = CONSTANTS.DIFF_EVENT_TYPE
        self._funding_info_messages_queue_key = CONSTANTS.FUNDING_EVENT_TYPE
        self._domain = domain
        self._api_factory = api_factory

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        ex_trading_pair = self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        params = {"symbol": ex_trading_pair}
        data = await self._connector._api_get(
            path_url=CONSTANTS.MARK_PRICE_PATH_URL,
            params=params,
            throttler_limit_id=CONSTANTS.MARK_PRICE_PATH_URL)
        return FundingInfo(trading_pair=trading_pair,
                           index_price=Decimal(data[0]["indexPrice"]),
                           mark_price=Decimal(data[0]["markPrice"]),
                           next_funding_utc_timestamp=data[0]["nextFundingTimestamp"] * 1e-3,
                           rate=Decimal(data[0]["fundingRate"]))

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        params = {
            "symbol": self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
            "limit": "1000"
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.SNAPSHOT_PATH_URL, domain=self._domain),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.SNAPSHOT_PATH_URL,
        )
        return data

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=CONSTANTS.WSS_URL.format(self._domain),
                         ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            for trading_pair in self._trading_pairs:
                trading_pair = self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                await self.subscribe_to_trading_pair(trading_pair)
                await self.subscribe_funding_info(trading_pair)
            self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            raise

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket not connected"
            )
            return False

        trade_params = [f"trade.{trading_pair}"]
        payload = {
            "method": "SUBSCRIBE",
            "params": trade_params,
        }
        subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=payload)

        depth_params = [f"depth.{trading_pair}"]
        payload = {
            "method": "SUBSCRIBE",
            "params": depth_params,
        }
        subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)

        try:
            await self._ws_assistant.send(subscribe_trade_request)
            await self._ws_assistant.send(subscribe_orderbook_request)
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error subscribing to {trading_pair}")
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket not connected"
            )
            return False

        trade_params = [f"trade.{trading_pair}"]
        payload = {
            "method": "UNSUBSCRIBE",
            "params": trade_params,
        }
        unsubscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=payload)

        depth_params = [f"depth.{trading_pair}"]
        payload = {
            "method": "UNSUBSCRIBE",
            "params": depth_params,
        }
        unsubscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)

        try:
            await self._ws_assistant.send(unsubscribe_trade_request)
            await self._ws_assistant.send(unsubscribe_orderbook_request)
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred unsubscribing from {trading_pair}...",
                exc_info=True
            )
            return False

    async def subscribe_funding_info(self, trading_pair: str) -> None:
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket not connected"
            )
            return

        funding_info_params = [f"markPrice.{trading_pair}"]
        payload = {
            "method": "SUBSCRIBE",
            "params": funding_info_params,
        }
        subscribe_funding_info_request: WSJSONRequest = WSJSONRequest(payload=payload)

        try:
            await self._ws_assistant.send(subscribe_funding_info_request)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(f"Unexpected error occurred subscribing to funding info for {trading_pair}...")

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        stream = event_message.get("stream", "")
        if CONSTANTS.DIFF_EVENT_TYPE in stream:
            channel = self._diff_messages_queue_key
        elif CONSTANTS.TRADE_EVENT_TYPE in stream:
            channel = self._trade_messages_queue_key
        elif CONSTANTS.FUNDING_EVENT_TYPE in stream:
            channel = self._funding_info_messages_queue_key
        return channel

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = time.time()
        snapshot_msg: OrderBookMessage = BackpackPerpetualOrderBook.snapshot_message_from_exchange(
            snapshot,
            snapshot_timestamp,
            metadata={"trading_pair": trading_pair}
        )
        return snapshot_msg

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if "data" in raw_message and CONSTANTS.DIFF_EVENT_TYPE in raw_message.get("stream"):
            trading_pair = self._connector.trading_pair_associated_to_exchange_symbol(symbol=raw_message["data"]["s"])
            order_book_message: OrderBookMessage = BackpackPerpetualOrderBook.diff_message_from_exchange(
                raw_message, time.time(), {"trading_pair": trading_pair})
            message_queue.put_nowait(order_book_message)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if "data" in raw_message and CONSTANTS.TRADE_EVENT_TYPE in raw_message.get("stream"):
            trading_pair = self._connector.trading_pair_associated_to_exchange_symbol(symbol=raw_message["data"]["s"])
            trade_message = BackpackPerpetualOrderBook.trade_message_from_exchange(
                raw_message, {"trading_pair": trading_pair})
            message_queue.put_nowait(trade_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue) -> None:
        data: Dict[str, Any] = raw_message["data"]
        trading_pair: str = self._connector.trading_pair_associated_to_exchange_symbol(data["s"])
        funding_update = FundingInfoUpdate(
            trading_pair=trading_pair,
            index_price=Decimal(data["i"]),
            mark_price=Decimal(data["p"]),
            next_funding_utc_timestamp=int(int(data["n"]) * 1e-3),
            rate=Decimal(data["f"])
        )
        message_queue.put_nowait(funding_update)



from decimal import Decimal
from typing import List

from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple

from .data_types import ArbProposal, ArbProposalSide

s_decimal_nan = Decimal("NaN")


async def create_arb_proposals(market_info_1: MarketTradingPairTuple,
                               market_info_2: MarketTradingPairTuple,
                               order_amount: Decimal) -> List[ArbProposal]:
    """
    Creates base arbitrage proposals for given markets without any filtering.
    :param market_info_1: The first market
    :param market_info_2: The second market
    :param order_amount: The required order amount.
    :return A list of 2 proposal - (market_1 buy, market_2 sell) and (market_1 sell, market_2 buy)
    """
    order_amount = Decimal(str(order_amount))
    results = []
    for index in range(0, 2):
        is_buy = not bool(index)  # bool(0) is False, so start with buy first
        m_1_q_price = await market_info_1.market.get_quote_price(market_info_1.trading_pair, is_buy, order_amount)
        m_1_o_price = await market_info_1.market.get_order_price(market_info_1.trading_pair, is_buy, order_amount)
        m_2_q_price = await market_info_2.market.get_quote_price(market_info_2.trading_pair, not is_buy, order_amount)
        m_2_o_price = await market_info_2.market.get_order_price(market_info_2.trading_pair, not is_buy, order_amount)
        if any(p is None for p in (m_1_o_price, m_1_q_price, m_2_o_price, m_2_q_price)):
            continue
        first_side = ArbProposalSide(
            market_info_1,
            is_buy,
            m_1_q_price,
            m_1_o_price,
            order_amount
        )
        second_side = ArbProposalSide(
            market_info_2,
            not is_buy,
            m_2_q_price,
            m_2_o_price,
            order_amount
        )
        results.append(ArbProposal(first_side, second_side))
    return results



from enum import Enum

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionMode
from hummingbot.core.data_type.in_flight_order import OrderState


class MarginMode(Enum):
    CROSS = "CROSS"
    ISOLATED = "ISOLATED"


EXCHANGE_NAME = "bitget_perpetual"
DEFAULT_DOMAIN = "bitget.com"
REST_SUBDOMAIN = "api"
WSS_SUBDOMAIN = "ws"
DEFAULT_TIME_IN_FORCE = "gtc"

ORDER_ID_MAX_LEN = None
HBOT_ORDER_ID_PREFIX = ""

WSS_PUBLIC_ENDPOINT = "/v2/ws/public"
WSS_PRIVATE_ENDPOINT = "/v2/ws/private"

MARGIN_MODE_TYPES = {
    MarginMode.CROSS: "crossed",
    MarginMode.ISOLATED: "isolated",
}
ORDER_TYPES = {
    OrderType.LIMIT: "limit",
    OrderType.MARKET: "market",
}
POSITION_MODE_TYPES = {
    PositionMode.ONEWAY: "one_way_mode",
    PositionMode.HEDGE: "hedge_mode",
}
STATE_TYPES = {
    "live": OrderState.OPEN,
    "filled": OrderState.FILLED,
    "partially_filled": OrderState.PARTIALLY_FILLED,
    "cancelled": OrderState.CANCELED,
    "canceled": OrderState.CANCELED,
}

SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 20
WS_HEARTBEAT_TIME_INTERVAL = 30

USDT_PRODUCT_TYPE = "USDT-FUTURES"
USDC_PRODUCT_TYPE = "USDC-FUTURES"
USD_PRODUCT_TYPE = "COIN-FUTURES"
ALL_PRODUCT_TYPES = [USDT_PRODUCT_TYPE, USDC_PRODUCT_TYPE, USD_PRODUCT_TYPE]

PUBLIC_TICKER_ENDPOINT = "/api/v2/mix/market/ticker"
PUBLIC_CONTRACTS_ENDPOINT = "/api/v2/mix/market/contracts"
PUBLIC_ORDERBOOK_ENDPOINT = "/api/v2/mix/market/merge-depth"
PUBLIC_FUNDING_RATE_ENDPOINT = "/api/v2/mix/market/current-fund-rate"
PUBLIC_OPEN_INTEREST_ENDPOINT = "/api/v2/mix/market/open-interest"
PUBLIC_SYMBOL_PRICE_ENDPOINT = "/api/v2/mix/market/symbol-price"
PUBLIC_TIME_ENDPOINT = "/api/v2/public/time"
PUBLIC_FUNDING_TIME_ENDPOINT = "/api/v2/mix/market/funding-time"

SET_LEVERAGE_ENDPOINT = "/api/v2/mix/account/set-leverage"
ALL_POSITIONS_ENDPOINT = "/api/v2/mix/position/all-position"
PLACE_ORDER_ENDPOINT = "/api/v2/mix/order/place-order"
CANCEL_ORDER_ENDPOINT = "/api/v2/mix/order/cancel-order"
ORDER_DETAIL_ENDPOINT = "/api/v2/mix/order/detail"
ORDER_FILLS_ENDPOINT = "/api/v2/mix/order/fills"
ACCOUNTS_INFO_ENDPOINT = "/api/v2/mix/account/accounts"
ACCOUNT_INFO_ENDPOINT = "/api/v2/mix/account/account"
SET_POSITION_MODE_ENDPOINT = "/api/v2/mix/account/set-position-mode"
SET_MARGIN_MODE_ENDPOINT = "/api/v2/mix/account/set-margin-mode"
ACCOUNT_BILLS_ENDPOINT = "/api/v2/mix/account/bill"

API_CODE = "bntva"

PUBLIC_WS_BOOKS = "books"
PUBLIC_WS_TRADE = "trade"
PUBLIC_WS_TICKER = "ticker"

PUBLIC_WS_PING_REQUEST = "ping"
PUBLIC_WS_PONG_RESPONSE = "pong"

WS_POSITIONS_ENDPOINT = "positions"
WS_ORDERS_ENDPOINT = "orders"
WS_ACCOUNT_ENDPOINT = "account"

RET_CODE_OK = "00000"
RET_CODE_PARAMS_ERROR = "40007"
RET_CODE_API_KEY_INVALID = "40006"
RET_CODE_AUTH_TIMESTAMP_ERROR = "40005"
RET_CODES_ORDER_NOT_EXISTS = [
    "40768", "80011", "40819",
    "43020", "43025", "43001",
    "45057", "31007", "43033"
]
RET_CODE_API_KEY_EXPIRED = "40014"


RATE_LIMITS = [
    RateLimit(limit_id=PUBLIC_TICKER_ENDPOINT, limit=20, time_interval=1),
    RateLimit(limit_id=PUBLIC_CONTRACTS_ENDPOINT, limit=20, time_interval=1),
    RateLimit(limit_id=PUBLIC_ORDERBOOK_ENDPOINT, limit=20, time_interval=1),
    RateLimit(limit_id=PUBLIC_TIME_ENDPOINT, limit=20, time_interval=1),
    RateLimit(limit_id=PUBLIC_FUNDING_RATE_ENDPOINT, limit=20, time_interval=1),
    RateLimit(limit_id=PUBLIC_OPEN_INTEREST_ENDPOINT, limit=20, time_interval=1),
    RateLimit(limit_id=PUBLIC_SYMBOL_PRICE_ENDPOINT, limit=20, time_interval=1),
    RateLimit(limit_id=PUBLIC_FUNDING_TIME_ENDPOINT, limit=20, time_interval=1),

    RateLimit(limit_id=SET_LEVERAGE_ENDPOINT, limit=5, time_interval=1),
    RateLimit(limit_id=ALL_POSITIONS_ENDPOINT, limit=5, time_interval=1),
    RateLimit(limit_id=PLACE_ORDER_ENDPOINT, limit=10, time_interval=1),
    RateLimit(limit_id=CANCEL_ORDER_ENDPOINT, limit=10, time_interval=1),
    RateLimit(limit_id=ORDER_DETAIL_ENDPOINT, limit=10, time_interval=1),
    RateLimit(limit_id=ORDER_FILLS_ENDPOINT, limit=10, time_interval=1),
    RateLimit(limit_id=ACCOUNTS_INFO_ENDPOINT, limit=10, time_interval=1),
    RateLimit(limit_id=ACCOUNT_INFO_ENDPOINT, limit=10, time_interval=1),
    RateLimit(limit_id=ACCOUNT_BILLS_ENDPOINT, limit=10, time_interval=1),
    RateLimit(limit_id=SET_POSITION_MODE_ENDPOINT, limit=5, time_interval=1),
    RateLimit(limit_id=SET_MARGIN_MODE_ENDPOINT, limit=5, time_interval=1),
]



import json
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict

from pydantic import Field, SecretStr

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.connector.exchange.foxbit import foxbit_constants as CONSTANTS
from hummingbot.core.data_type.in_flight_order import OrderState
from hummingbot.core.data_type.trade_fee import TradeFeeSchema
from hummingbot.core.utils.tracking_nonce import get_tracking_nonce

CENTRALIZED = True
EXAMPLE_PAIR = "BTC-BRL"
_seq_nr: int = 0

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.001"),
    taker_percent_fee_decimal=Decimal("0.001"),
    buy_percent_fee_deducted_from_returns=True
)


def get_client_order_id(is_buy: bool) -> str:
    """
    Creates a client order id for a new order
    :param is_buy: True if the order is a buy order, False if the order is a sell order
    :return: an identifier for the new order to be used in the client
    """
    newId = str(get_tracking_nonce())[4:]
    side = "00" if is_buy else "01"
    return f"{CONSTANTS.HBOT_ORDER_ID_PREFIX}{side}{newId}"


def get_ws_message_frame(endpoint: str,
                         msg_type: str = "0",
                         payload: str = "",
                         ) -> Dict[str, Any]:
    retValue = CONSTANTS.WS_MESSAGE_FRAME.copy()
    retValue["m"] = msg_type
    retValue["i"] = _get_next_message_frame_sequence_number()
    retValue["n"] = endpoint
    retValue["o"] = json.dumps(payload)
    return retValue


def _get_next_message_frame_sequence_number() -> int:
    """
    Returns next sequence number to be used into message frame for WS requests
    """
    global _seq_nr
    _seq_nr += 1
    return _seq_nr


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair. Dictionary with status and permissions
    :return: True if the trading pair is enabled, False otherwise

    Nowadays all available pairs are valid.
    It is here for future implamentation.
    """
    return True


def ws_data_to_dict(data: str) -> Dict[str, Any]:
    return eval(data.replace(":null", ":None").replace(":false", ":False").replace(":true", ":True"))


def datetime_val_or_now(string_value: str,
                        string_format: str = '%Y-%m-%dT%H:%M:%S.%fZ',
                        on_error_return_now: bool = True,
                        ) -> datetime:
    try:
        return datetime.strptime(string_value, string_format)
    except Exception:
        if on_error_return_now:
            return datetime.now()
        else:
            return None


def decimal_val_or_none(string_value: str,
                        on_error_return_none: bool = True,
                        ) -> Decimal:
    try:
        return Decimal(string_value)
    except Exception:
        if on_error_return_none:
            return None
        else:
            return Decimal('0')


def int_val_or_none(string_value: str,
                    on_error_return_none: bool = True,
                    ) -> int:
    try:
        return int(string_value)
    except Exception:
        if on_error_return_none:
            return None
        else:
            return int('0')


def get_order_state(state: str,
                    on_error_return_failed: bool = False,
                    ) -> OrderState:
    try:
        return CONSTANTS.ORDER_STATE[state]
    except Exception:
        if on_error_return_failed:
            return OrderState.FAILED
        else:
            return None


def get_base_quote_from_trading_pair(trading_pair: str):
    if len(trading_pair) == 0:
        return "", ""
    if trading_pair.find("-") == -1:
        return "", ""
    pair = trading_pair.split("-")
    return pair[0].upper(), pair[1].upper()


class FoxbitConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="foxbit", client_data=None)
    foxbit_api_key: SecretStr = Field(
        default=...,
        json_schema_extra = {
            "prompt": "Enter your Foxbit API key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    foxbit_api_secret: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Foxbit API secret",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    foxbit_user_id: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: "Enter your Foxbit User ID",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )

    class Config:
        title = "foxbit"


KEYS = FoxbitConfigMap.construct()

OTHER_DOMAINS = []
OTHER_DOMAINS_PARAMETER = {}
OTHER_DOMAINS_EXAMPLE_PAIR = {}
OTHER_DOMAINS_DEFAULT_FEES = {}
OTHER_DOMAINS_KEYS = {}



import asyncio
import logging
from typing import TYPE_CHECKING, Any, Dict, Optional

from hummingbot.connector.exchange.ndax import ndax_constants as CONSTANTS
from hummingbot.connector.exchange.ndax.ndax_auth import NdaxAuth
from hummingbot.connector.exchange.ndax.ndax_websocket_adaptor import NdaxWebSocketAdaptor
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.ndax.ndax_exchange import NdaxExchange


class NdaxAPIUserStreamDataSource(UserStreamTrackerDataSource):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(
        self,
        auth: NdaxAuth,
        trading_pairs: str,
        connector: "NdaxExchange",
        api_factory: WebAssistantsFactory,
        domain: Optional[str] = None,
    ):
        super().__init__()
        self._trading_pairs = trading_pairs
        self._ws_adaptor = None
        self._auth_assistant: NdaxAuth = auth
        self._last_recv_time: float = 0
        self._account_id: Optional[int] = None
        self._oms_id: Optional[int] = None
        self._domain = domain
        self._api_factory = api_factory
        self._connector = connector

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    async def _connected_websocket_assistant(self) -> NdaxWebSocketAdaptor:
        """
        Creates an instance of WSAssistant connected to the exchange
        """
        ws: WSAssistant = await self._get_ws_assistant()
        url = CONSTANTS.WSS_URLS.get(self._domain or "ndax_main")
        await ws.connect(ws_url=url)
        return NdaxWebSocketAdaptor(ws)

    async def _authenticate(self, ws: NdaxWebSocketAdaptor):
        """
        Authenticates user to websocket
        """
        try:
            await ws.send_request(
                CONSTANTS.AUTHENTICATE_USER_ENDPOINT_NAME, self._auth_assistant.header_for_authentication()
            )
            auth_resp = await ws.websocket.receive()
            auth_payload: Dict[str, Any] = ws.payload_from_raw_message(auth_resp.data)

            if not auth_payload["Authenticated"]:
                self.logger().error(f"Response: {auth_payload}", exc_info=True)
                raise Exception("Could not authenticate websocket connection with NDAX")

            auth_user = auth_payload.get("User")
            self._account_id = auth_user.get("AccountId")
            self._oms_id = auth_user.get("OMSId")

        except asyncio.CancelledError:
            raise
        except Exception as ex:
            self.logger().error(f"Error occurred when authenticating to user stream ({ex})", exc_info=True)
            raise

    async def _subscribe_channels(self, ws: NdaxWebSocketAdaptor):
        """
        Subscribes to User Account Events
        """
        payload = {"AccountId": self._account_id, "OMSId": self._oms_id}
        try:
            await ws.send_request(CONSTANTS.SUBSCRIBE_ACCOUNT_EVENTS_ENDPOINT_NAME, payload)
        except asyncio.CancelledError:
            raise
        except Exception as ex:
            self.logger().error(
                f"Error occurred subscribing to {CONSTANTS.EXCHANGE_NAME} private channels ({ex})", exc_info=True
            )
            raise

    async def listen_for_user_stream(self, output: asyncio.Queue):
        """
        *required
        Subscribe to user stream via web socket, and keep the connection open for incoming messages

        :param output: an async queue where the incoming messages are stored
        """
        while True:
            try:
                ws: NdaxWebSocketAdaptor = await self._connected_websocket_assistant()
                self.logger().info("Authenticating to User Stream...")
                await self._authenticate(ws)
                self.logger().info("Successfully authenticated to User Stream.")
                await self._subscribe_channels(ws)
                self.logger().info("Successfully subscribed to user events.")

                await ws.process_websocket_messages(queue=output)
            except asyncio.CancelledError:
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(f"The websocket connection was closed ({connection_exception})")
            except Exception:
                self.logger().exception("Unexpected error while listening to user stream. Retrying after 5 seconds...")
                await self._sleep(1.0)
            finally:
                await self._on_user_stream_interruption(websocket_assistant=self._ws_assistant)
                self._ws_assistant = None



import json
import time

import base58
from solders.keypair import Keypair

from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class PacificaPerpetualAuth(AuthBase):
    def __init__(self, agent_wallet_public_key: str, agent_wallet_private_key: str, user_wallet_public_key: str):
        # aka "agent_wallet"; we pass it in POST requests
        self.agent_wallet_public_key = agent_wallet_public_key
        # used to generate signature for POST requests
        self.agent_wallet_private_key = agent_wallet_private_key

        # aka "account"; we pass it to some GET requests and to all POST requests
        self.user_wallet_public_key = user_wallet_public_key

        self._keypair = None

    @property
    def keypair(self):
        if self._keypair is None:
            self._keypair = Keypair.from_bytes(base58.b58decode(self.agent_wallet_private_key))
        return self._keypair

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        if request.method == RESTMethod.POST:
            request_data = json.loads(request.data)

            operation_type: str = request_data.pop("type")

            signature_header = {
                "timestamp": int(time.time() * 1000),
                "expiry_window": 5000,
                "type": operation_type,
            }

            _, signature_b58 = sign_message(signature_header, request_data, self.keypair)

            final_body = {
                "account": self.user_wallet_public_key,
                "agent_wallet": self.agent_wallet_public_key,
                "signature": signature_b58,
                "timestamp": signature_header["timestamp"],
                "expiry_window": signature_header["expiry_window"],
                **request_data
            }

            request.data = json.dumps(final_body)
            request.headers = {"Content-Type": "application/json"}

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        params = request.payload.get("params", {})

        if params is None:
            return request

        operation_type: str = params.pop("type")

        signature_header = {
            "timestamp": int(time.time() * 1000),
            "expiry_window": 5000,
            "type": operation_type,
        }

        _, signature_b58 = sign_message(signature_header, params, self.keypair)

        final_body = {
            "account": self.agent_wallet_public_key,
            "agent_wallet": self.agent_wallet_public_key,
            "signature": signature_b58,
            "timestamp": signature_header["timestamp"],
            "expiry_window": signature_header["expiry_window"],
            **params
        }

        request.payload["params"] = final_body

        return request

# the following 3 functions have been extracted from the official SDK
# https://github.com/pacifica-fi/python-sdk


def sign_message(header, payload, keypair):
    message = prepare_message(header, payload)
    message_bytes = message.encode("utf-8")
    signature = keypair.sign_message(message_bytes)
    return (message, base58.b58encode(bytes(signature)).decode("ascii"))


def sort_json_keys(value):
    if isinstance(value, dict):
        sorted_dict = {}
        for key in sorted(value.keys()):
            sorted_dict[key] = sort_json_keys(value[key])
        return sorted_dict
    elif isinstance(value, list):
        return [sort_json_keys(item) for item in value]
    else:
        return value


def prepare_message(header, payload):
    if (
        "type" not in header
        or "timestamp" not in header
        or "expiry_window" not in header
    ):
        raise ValueError("Header must have type, timestamp, and expiry_window")

    data = {
        **header,
        "data": payload,
    }

    message = sort_json_keys(data)

    # Specifying the separaters is important because the JSON message is expected to be compact.
    message = json.dumps(message, separators=(",", ":"))

    return message



# distutils: language=c++

from libc.stdint cimport int64_t
from hummingbot.strategy.__utils__.trailing_indicators.trading_intensity cimport TradingIntensityIndicator
from hummingbot.strategy.strategy_base cimport StrategyBase


cdef class AvellanedaMarketMakingStrategy(StrategyBase):
    cdef:
        object _config_map
        object _market_info
        object _price_delegate
        object _minimum_spread
        bint _hanging_orders_enabled
        object _hanging_orders_cancel_pct
        object _hanging_orders_tracker
        bint _add_transaction_costs_to_orders
        bint _hb_app_notification
        bint _is_debug

        double _cancel_timestamp
        double _create_timestamp
        object _limit_order_type
        bint _all_markets_ready
        int _filled_buys_balance
        int _filled_sells_balance
        double _last_timestamp
        double _status_report_interval
        int64_t _logging_options
        object _last_own_trade_price
        int _volatility_sampling_period
        double _last_sampling_timestamp
        bint _parameters_based_on_spread
        int _volatility_buffer_size
        int _trading_intensity_buffer_size
        int _ticks_to_be_ready
        object _alpha
        object _kappa
        object _gamma
        object _eta
        str _execution_mode
        str _execution_timeframe
        object _execution_state
        object _start_time
        object _end_time
        double _min_spread
        object _q_adjustment_factor
        object _reservation_price
        object _optimal_spread
        object _optimal_bid
        object _optimal_ask
        str _debug_csv_path
        object _avg_vol
        TradingIntensityIndicator _trading_intensity
        bint _should_wait_order_cancel_confirmation

    cdef object c_get_mid_price(self)
    cdef _create_proposal_based_on_order_levels(self)
    cdef _create_proposal_based_on_order_override(self)
    cdef _create_basic_proposal(self)
    cdef object c_create_base_proposal(self)
    cdef tuple c_get_adjusted_available_balance(self, list orders)
    cdef c_apply_order_price_modifiers(self, object proposal)
    cdef c_apply_order_amount_eta_transformation(self, object proposal)
    cdef c_apply_budget_constraint(self, object proposal)
    cdef c_apply_order_optimization(self, object proposal)
    cdef c_apply_add_transaction_costs(self, object proposal)
    cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices)
    cdef c_cancel_active_orders(self, object proposal)
    cdef c_cancel_active_orders_on_max_age_limit(self)
    cdef bint c_to_create_orders(self, object proposal)
    cdef c_execute_orders_proposal(self, object proposal)
    cdef c_set_timers(self)
    cdef double c_get_spread(self)
    cdef c_collect_market_variables(self, double timestamp)
    cdef bint c_is_algorithm_ready(self)
    cdef bint c_is_algorithm_changed(self)
    cdef c_measure_order_book_liquidity(self)
    cdef c_calculate_reservation_price_and_optimal_spread(self)
    cdef object c_calculate_target_inventory(self)
    cdef object c_calculate_inventory(self)
    cdef c_did_complete_order(self, object order_completed_event)



# A single source of truth for constant variables related to the exchange
from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

# Max order id allowed by AscendEx is 32. We need to configure it in 22 to unify the timestamp and client_id
# components in the order id, because AscendEx uses the last 9 characters or the order id to generate the
# exchange order id (the last 9 characters would be part of the client id if it is not mixed with the timestamp).
# But AscendEx only uses milliseconds for the timestamp in the exchange order id, causing duplicated exchange order ids
# when running strategies with multi level orders created very fast.
MAX_ORDER_ID_LEN = 22

PING_TIMEOUT = 15.0
DEFAULT_DOMAIN = ""
HBOT_ORDER_ID_PREFIX = "HMBot"

EXCHANGE_NAME = "ascend_ex"
PUBLIC_REST_URL = "https://ascendex.com/api/pro/v1/"
PRIVATE_REST_URL = "https://ascendex.com/{group_id}/api/pro/v1/"
WS_URL = "wss://ascendex.com:443/api/pro/v1/websocket-for-hummingbot-liq-mining"
PRIVATE_WS_URL = "wss://ascendex.com:443/{group_id}/api/pro/v1/websocket-for-hummingbot-liq-mining"

# REST API ENDPOINTS
ORDER_PATH_URL = "cash/order"
ORDER_BATCH_PATH_URL = "cash/order/batch"
ORDER_OPEN_PATH_URL = "cash/order/open"
ORDER_STATUS_PATH_URL = "cash/order/status"
BALANCE_PATH_URL = "cash/balance"
BALANCE_HISTORY_PATH_URL = "data/v1/cash/balance/history"
HIST_PATH_URL = "cash/order/hist/current"
FEE_PATH_URL = "spot/fee"
TICKER_PATH_URL = "spot/ticker"
PRODUCTS_PATH_URL = "cash/products"
TRADES_PATH_URL = "trades"
DEPTH_PATH_URL = "depth"
INFO_PATH_URL = "info"
STREAM_PATH_URL = "stream"

SERVER_LIMIT_INFO = "risk-limit-info"

# WS API ENDPOINTS
SUB_ENDPOINT_NAME = "sub"
PONG_ENDPOINT_NAME = "pong"
TRADE_TOPIC_ID = "trades"
DIFF_TOPIC_ID = "depth"
PING_TOPIC_ID = "ping"
ACCOUNT_TYPE = "CASH"
BALANCE_EVENT_TYPE = "balance"
ORDER_CHANGE_EVENT_TYPE = "order"

# OrderStates

ORDER_STATE = {
    "PendingNew": OrderState.PENDING_CREATE,
    "New": OrderState.OPEN,
    "Filled": OrderState.FILLED,
    "PartiallyFilled": OrderState.PARTIALLY_FILLED,
    "Canceled": OrderState.CANCELED,
    "Rejected": OrderState.FAILED,
}

# AscendEx has multiple pools for API request limits
# Any call increases call rate in ALL pool, so e.g. a cash/order call will contribute to both ALL and cash/order pools.
ALL_ENDPOINTS_LIMIT = "All"
RATE_LIMITS = [
    RateLimit(limit_id=ALL_ENDPOINTS_LIMIT, limit=100, time_interval=1),
    RateLimit(
        limit_id=ORDER_PATH_URL, limit=50, time_interval=1, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=SERVER_LIMIT_INFO,
        limit=50,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=ORDER_BATCH_PATH_URL,
        limit=50,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=ORDER_OPEN_PATH_URL,
        limit=50,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=ORDER_STATUS_PATH_URL,
        limit=50,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=BALANCE_PATH_URL,
        limit=100,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=BALANCE_HISTORY_PATH_URL,
        limit=8,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=HIST_PATH_URL, limit=60, time_interval=60, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=TICKER_PATH_URL, limit=100, time_interval=1, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=PRODUCTS_PATH_URL,
        limit=100,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=FEE_PATH_URL, limit=100, time_interval=1, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=TRADES_PATH_URL, limit=100, time_interval=1, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=DEPTH_PATH_URL, limit=100, time_interval=1, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=INFO_PATH_URL, limit=100, time_interval=1, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=SUB_ENDPOINT_NAME,
        limit=100,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=PONG_ENDPOINT_NAME,
        limit=100,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
]



from decimal import Decimal
from enum import Enum
from typing import Literal, Optional

from pydantic import BaseModel, field_validator
from pydantic_core.core_schema import ValidationInfo

from hummingbot.core.data_type.common import PositionAction, TradeType
from hummingbot.strategy_v2.executors.data_types import ExecutorConfigBase


class ExecutionStrategy(Enum):
    LIMIT = "LIMIT"
    LIMIT_MAKER = "LIMIT_MAKER"
    MARKET = "MARKET"
    LIMIT_CHASER = "LIMIT_CHASER"


class LimitChaserConfig(BaseModel):
    distance: Decimal
    refresh_threshold: Decimal


class OrderExecutorConfig(ExecutorConfigBase):
    type: Literal["order_executor"] = "order_executor"
    trading_pair: str
    connector_name: str
    side: TradeType
    amount: Decimal
    position_action: PositionAction = PositionAction.OPEN
    price: Optional[Decimal] = None  # Required for LIMIT and LIMIT_MAKER
    chaser_config: Optional[LimitChaserConfig] = None  # Required for LIMIT_CHASER
    execution_strategy: ExecutionStrategy
    leverage: int = 1
    level_id: Optional[str] = None

    @field_validator("execution_strategy", mode="before")
    @classmethod
    def validate_execution_strategy(cls, value, validation_info: ValidationInfo):
        if value in [ExecutionStrategy.LIMIT, ExecutionStrategy.LIMIT_MAKER]:
            if validation_info.data.get('price') is None:
                raise ValueError("Price is required for LIMIT and LIMIT_MAKER execution strategies")
        elif value == ExecutionStrategy.LIMIT_CHASER:
            if validation_info.data.get('chaser_config') is None:
                raise ValueError("Chaser config is required for LIMIT_CHASER execution strategy")
        return value



import asyncio
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.bitstamp import bitstamp_constants as CONSTANTS, bitstamp_web_utils as web_utils
from hummingbot.connector.exchange.bitstamp.bitstamp_order_book import BitstampOrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.bitstamp.bitstamp_exchange import BitstampExchange


class BitstampAPIOrderBookDataSource(OrderBookTrackerDataSource):
    _logger: Optional[HummingbotLogger] = None
    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'BitstampExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__(trading_pairs)
        self._connector = connector
        self._trade_messages_queue_key = CONSTANTS.TRADE_EVENT_TYPE
        self._diff_messages_queue_key = CONSTANTS.DIFF_EVENT_TYPE
        self._domain = domain
        self._api_factory = api_factory
        self._channel_associated_to_tradingpair = {}

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

        rest_assistant = await self._api_factory.get_rest_assistant()
        order_book_data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.ORDER_BOOK_URL.format(symbol), domain=self._domain),
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.ORDER_BOOK_URL_LIMIT_ID,
        )

        return order_book_data

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            self._channel_associated_to_tradingpair.clear()
            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

                channel = CONSTANTS.WS_PUBLIC_LIVE_TRADES.format(symbol)
                payload = {
                    "event": "bts:subscribe",
                    "data": {
                        "channel": channel
                    }
                }
                subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=payload)
                self._channel_associated_to_tradingpair[channel] = trading_pair

                channel = CONSTANTS.WS_PUBLIC_DIFF_ORDER_BOOK.format(symbol)
                payload = {
                    "event": "bts:subscribe",
                    "data": {
                        "channel": channel
                    }
                }
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
                self._channel_associated_to_tradingpair[channel] = trading_pair

                await ws.send(subscribe_trade_request)
                await ws.send(subscribe_orderbook_request)

            self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            raise

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange
        """
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=CONSTANTS.WSS_URL.format(self._domain),
                         ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_msg: OrderBookMessage = BitstampOrderBook.snapshot_message_from_exchange(
            snapshot,
            time.time(),
            metadata={"trading_pair": trading_pair}
        )
        return snapshot_msg

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = self._channel_associated_to_tradingpair.get(raw_message["channel"])

        trade_message = BitstampOrderBook.trade_message_from_exchange(
            raw_message, {"trading_pair": trading_pair})

        message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = self._channel_associated_to_tradingpair.get(raw_message["channel"])

        order_book_message: OrderBookMessage = BitstampOrderBook.diff_message_from_exchange(
            raw_message, time.time(), {"trading_pair": trading_pair})

        message_queue.put_nowait(order_book_message)

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        return event_message.get("event", "")

    async def _process_message_for_unknown_channel(self, event_message: Dict[str, Any], websocket_assistant: WSAssistant):
        event = event_message.get("event", "")
        channel = event_message.get("channel")
        if event == "bts:subscription_succeeded":
            self.logger().info(f"Subscription succeeded for channel '{channel}'")
        elif event == "bts:request_reconnect":
            raise ConnectionError("Received request to reconnect. Reconnecting...")
        else:
            self.logger().debug(f"Received message from unknown channel: {event_message}")

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        subscribe_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return subscribe_id

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribe to order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to subscribe to
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot subscribe: WebSocket connection not established")
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            channel = CONSTANTS.WS_PUBLIC_LIVE_TRADES.format(symbol)
            payload = {
                "event": "bts:subscribe",
                "data": {
                    "channel": channel
                }
            }
            subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=payload)
            self._channel_associated_to_tradingpair[channel] = trading_pair

            channel = CONSTANTS.WS_PUBLIC_DIFF_ORDER_BOOK.format(symbol)
            payload = {
                "event": "bts:subscribe",
                "data": {
                    "channel": channel
                }
            }
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
            self._channel_associated_to_tradingpair[channel] = trading_pair

            await self._ws_assistant.send(subscribe_trade_request)
            await self._ws_assistant.send(subscribe_orderbook_request)

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred subscribing to {trading_pair}...",
                exc_info=True
            )
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribe from order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot unsubscribe: WebSocket connection not established")
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            trade_channel = CONSTANTS.WS_PUBLIC_LIVE_TRADES.format(symbol)
            payload = {
                "event": "bts:unsubscribe",
                "data": {
                    "channel": trade_channel
                }
            }
            unsubscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=payload)

            orderbook_channel = CONSTANTS.WS_PUBLIC_DIFF_ORDER_BOOK.format(symbol)
            payload = {
                "event": "bts:unsubscribe",
                "data": {
                    "channel": orderbook_channel
                }
            }
            unsubscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)

            await self._ws_assistant.send(unsubscribe_trade_request)
            await self._ws_assistant.send(unsubscribe_orderbook_request)

            # Remove channel mappings
            self._channel_associated_to_tradingpair.pop(trade_channel, None)
            self._channel_associated_to_tradingpair.pop(orderbook_channel, None)

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred unsubscribing from {trading_pair}...",
                exc_info=True
            )
            return False



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "pacifica_perpetual"
DEFAULT_DOMAIN = "pacifica_perpetual"
HB_OT_ID_PREFIX = "HBOT"

# Base URLs
REST_URL = "https://api.pacifica.fi/api/v1"
WSS_URL = "wss://ws.pacifica.fi/ws"

TESTNET_DOMAIN = "pacifica_perpetual_testnet"
TESTNET_REST_URL = "https://test-api.pacifica.fi/api/v1"
TESTNET_WSS_URL = "wss://test-ws.pacifica.fi/ws"

# order status mapping
ORDER_STATE = {
    "open": OrderState.OPEN,
    "filled": OrderState.FILLED,
    "partially_filled": OrderState.PARTIALLY_FILLED,
    "cancelled": OrderState.CANCELED,
    "rejected": OrderState.FAILED,
}

GET_MARKET_ORDER_BOOK_SNAPSHOT_PATH_URL = "/book"
GET_ORDER_HISTORY_PATH_URL = "/orders/history_by_id"
GET_CANDLES_PATH_URL = "/kline"
GET_PRICES_PATH_URL = "/info/prices"
GET_POSITIONS_PATH_URL = "/positions"
GET_FUNDING_HISTORY_PATH_URL = "/funding/history"
SET_LEVERAGE_PATH_URL = "/account/leverage"
CANCEL_ORDER_PATH_URL = "/orders/cancel"
EXCHANGE_INFO_PATH_URL = "/info"
GET_ACCOUNT_INFO_PATH_URL = "/account"
GET_ACCOUNT_API_CONFIG_KEYS = "/account/api_keys"
CREATE_ACCOUNT_API_CONFIG_KEY = "/account/api_keys/create"
GET_TRADE_HISTORY_PATH_URL = "/trades/history"
GET_FEES_INFO_PATH_URL = "/info/fees"

# the API endpoints for market / limit / stop orders are different
# the support for stop orders is out of the scope for this integration
CREATE_MARKET_ORDER_PATH_URL = "/orders/create_market"
CREATE_LIMIT_ORDER_PATH_URL = "/orders/create"

# Default maximum slippage tolerance for market orders (percentage string, e.g. "5" = 5%)
MARKET_ORDER_MAX_SLIPPAGE = "5"

# WebSocket Channels

WS_ORDER_BOOK_SNAPSHOT_CHANNEL = "book"
WS_TRADES_CHANNEL = "trades"
WS_PRICES_CHANNEL = "prices"

WS_ACCOUNT_ORDER_UPDATES_CHANNEL = "account_order_updates"
WS_ACCOUNT_POSITIONS_CHANNEL = "account_positions"
WS_ACCOUNT_INFO_CHANNEL = "account_info"
WS_ACCOUNT_TRADES_CHANNEL = "account_trades"

WS_PING_INTERVAL = 30  # Keep connection alive

# the exchange has different "costs" of the calls for every endpoint
# plus there're exactly 2 tiers of rate limits: (1) Unidentified IP (2) Valid API Config Key
# below you could find (in the comments) -- the costs (aka "weight") of each endpoints group

PACIFICA_LIMIT_ID = "PACIFICA_LIMIT"

# All values are x10 of doc values to support fractional costs (cancellation = 0.5 credits)
# since Hummingbot's throttler requires integer weights
PACIFICA_TIER_1_LIMIT = 1250  # Unidentified IP (doc: 125)
PACIFICA_TIER_2_LIMIT = 3000  # Valid API Config Key (doc: 300)
PACIFICA_LIMIT_INTERVAL = 60

FEE_TIER_LIMITS = {
    0: 3000,     # doc: 300
    1: 6000,     # doc: 600
    2: 12000,    # doc: 1200
    3: 24000,    # doc: 2400
    4: 60000,    # doc: 6000
    5: 120000,   # doc: 12000
    6: 240000,   # doc: 24000
    7: 300000,   # doc: 30000
}

# Costs (x10 of doc values)
STANDARD_REQUEST_COST = 10       # doc: 1
ORDER_CANCELLATION_COST = 5      # doc: 0.5
HEAVY_GET_REQUEST_COST_TIER_1 = 120  # Unidentified IP (doc: 12)
HEAVY_GET_REQUEST_COST_TIER_2 = 30   # Valid API Config Key (doc: 3)

RATE_LIMITS = [
    RateLimit(limit_id=PACIFICA_LIMIT_ID, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL),
    RateLimit(limit_id=GET_MARKET_ORDER_BOOK_SNAPSHOT_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=CREATE_LIMIT_ORDER_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=STANDARD_REQUEST_COST)]),
    RateLimit(limit_id=CREATE_MARKET_ORDER_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=STANDARD_REQUEST_COST)]),
    RateLimit(limit_id=CANCEL_ORDER_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=ORDER_CANCELLATION_COST)]),
    RateLimit(limit_id=SET_LEVERAGE_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=STANDARD_REQUEST_COST)]),
    RateLimit(limit_id=GET_FUNDING_HISTORY_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_POSITIONS_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_ORDER_HISTORY_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_CANDLES_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=EXCHANGE_INFO_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_PRICES_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_ACCOUNT_INFO_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_ACCOUNT_API_CONFIG_KEYS, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=CREATE_ACCOUNT_API_CONFIG_KEY, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_TRADE_HISTORY_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
    RateLimit(limit_id=GET_FEES_INFO_PATH_URL, limit=PACIFICA_TIER_1_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_1)]),
]

RATE_LIMITS_TIER_2 = [
    RateLimit(limit_id=PACIFICA_LIMIT_ID, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL),
    RateLimit(limit_id=GET_MARKET_ORDER_BOOK_SNAPSHOT_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=CREATE_LIMIT_ORDER_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=STANDARD_REQUEST_COST)]),
    RateLimit(limit_id=CREATE_MARKET_ORDER_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=STANDARD_REQUEST_COST)]),
    RateLimit(limit_id=CANCEL_ORDER_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=ORDER_CANCELLATION_COST)]),
    RateLimit(limit_id=SET_LEVERAGE_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=STANDARD_REQUEST_COST)]),
    RateLimit(limit_id=GET_FUNDING_HISTORY_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_POSITIONS_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_ORDER_HISTORY_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_CANDLES_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=EXCHANGE_INFO_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_PRICES_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_ACCOUNT_INFO_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_ACCOUNT_API_CONFIG_KEYS, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=CREATE_ACCOUNT_API_CONFIG_KEY, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_TRADE_HISTORY_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
    RateLimit(limit_id=GET_FEES_INFO_PATH_URL, limit=PACIFICA_TIER_2_LIMIT, time_interval=PACIFICA_LIMIT_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(limit_id=PACIFICA_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST_TIER_2)]),
]



# distutils: language=c++
# distutils: sources=hummingbot/core/cpp/OrderExpirationEntry.cpp

from libcpp.string cimport string
import pandas as pd
from typing import List

cdef class OrderExpirationEntry:
    @classmethod
    def to_pandas(cls, order_expiration_entries: List[OrderExpirationEntry]) -> pd.DataFrame:
        cdef:
            list columns = ["trading_pair", "order_id", "timestamp", "expiration", "expiration_time"]
            list data = [[
                expiration_entry.trading_pair,
                expiration_entry.order_id,
                expiration_entry.timestamp,
                expiration_entry.expiration,
                expiration_entry.expiration_time
            ] for expiration_entry in order_expiration_entries]

        return pd.DataFrame(data=data, columns=columns)

    def __init__(self,
                 trading_pair: str,
                 order_id: str,
                 timestamp: float,
                 expiration_ts: float
                 ):
        cdef:
            string cpp_trading_pair = trading_pair.encode("utf8")
            string cpp_order_id = order_id.encode("utf8")
            double cpp_timestamp = timestamp
            double cpp_expiration_ts = expiration_ts

        self._cpp_order_expiration_entry = CPPOrderExpirationEntry(cpp_trading_pair,
                                                                   cpp_order_id,
                                                                   cpp_timestamp,
                                                                   cpp_expiration_ts)

    @property
    def trading_pair(self) -> str:
        cdef:
            string cpp_trading_pair = self._cpp_order_expiration_entry.getTradingPair()
            str retval = cpp_trading_pair.decode("utf8")
        return retval

    @property
    def order_id(self) -> str:
        cdef:
            string cpp_client_order_id = self._cpp_order_expiration_entry.getClientOrderID()
            str retval = cpp_client_order_id.decode("utf8")
        return retval

    @property
    def timestamp(self) -> float:
        return self._cpp_order_expiration_entry.getTimestamp()

    @property
    def expiration_timestamp(self) -> float:
        return self._cpp_order_expiration_entry.getExpirationTimestamp()

    def __repr__(self) -> str:
        return f"OrderExpirationEntry('{self.trading_pair}', '{self.order_id}', {self.timestamp}, '{self.expiration_timestamp}')"


cdef OrderExpirationEntry c_create_order_expiration_from_cpp_order_expiration(const CPPOrderExpirationEntry cpp_order_expiration_entry):
    cdef OrderExpirationEntry retval = OrderExpirationEntry.__new__(OrderExpirationEntry)
    retval._cpp_order_expiration_entry = cpp_order_expiration_entry
    return retval



# distutils: language=c++

cdef extern from "cpp/PyRef.h":
    ctypedef struct PyObject

    cdef cppclass PyRef:
        PyRef()
        PyRef(PyObject *obj)
        PyRef(const PyRef &other)
        PyRef &operator=(const PyRef &other)
        bint operator==(const PyRef &other) const
        PyObject *get() const



import time
from typing import Any, Dict, Optional

import hummingbot.connector.derivative.gate_io_perpetual.gate_io_perpetual_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def public_rest_url(endpoint: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public REST endpoint

    :param endpoint: a public REST endpoint
    :param domain: unused

    :return: the full URL to the endpoint
    """
    if CONSTANTS.REST_URL[-1] != "/" and endpoint[0] != "/":
        endpoint = "/" + endpoint
    return CONSTANTS.REST_URL + endpoint


def private_rest_url(endpoint: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    return public_rest_url(endpoint, domain)


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        auth: Optional[AuthBase] = None) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth)
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(
        throttler: Optional[AsyncThrottler] = None,
        domain: str = CONSTANTS.DEFAULT_DOMAIN,
) -> float:
    return time.time()


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair
    :return: True if the trading pair is enabled, False otherwise
    """
    return exchange_info.get("in_delisting", None) is False



# distutils: language=c++

from libcpp cimport bool as cppbool
from libcpp.string cimport string

cdef extern from "../cpp/LimitOrder.h":
    ctypedef struct PyObject

    cdef cppclass LimitOrder:
        LimitOrder()
        LimitOrder(string clientOrderID,
                   string tradingPair,
                   cppbool isBuy,
                   string baseCurrency,
                   string quoteCurrency,
                   PyObject *price,
                   PyObject *quantity)
        LimitOrder(string clientOrderID,
                   string tradingPair,
                   cppbool isBuy,
                   string baseCurrency,
                   string quoteCurrency,
                   PyObject *price,
                   PyObject *quantity,
                   PyObject *filledQuantity,
                   long long creationTimestamp,
                   short int status,
                   string position)
        LimitOrder(const LimitOrder &other)
        LimitOrder &operator=(const LimitOrder &other)
        string getClientOrderID()
        string getTradingPair()
        cppbool getIsBuy()
        string getBaseCurrency()
        string getQuoteCurrency()
        PyObject *getPrice()
        PyObject *getQuantity()
        PyObject *getFilledQuantity()
        long long getCreationTimestamp()
        short int getStatus()
        string getPosition()



from typing import Any, Dict

import hummingbot.connector.derivative.dydx_v4_perpetual.dydx_v4_perpetual_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class DydxV4PerpetualRESTPreProcessor(RESTPreProcessorBase):

    async def pre_process(self, request: RESTRequest) -> RESTRequest:
        if request.headers is None:
            request.headers = {}
        request.headers["Accept"] = (
            "application/json"
        )
        return request


def public_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public REST endpoint
    :param path_url: a public REST endpoint
    :param domain: the dydx_v4 domain to connect to ("exchange" or "us"). The default value is "exchange"
    :return: the full URL to the endpoint
    """
    return CONSTANTS.DYDX_V4_REST_URL + path_url


def private_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided private REST endpoint
    :param path_url: a private REST endpoint
    :param domain: the dYdX domain to connect to ("exchange" or "us"). The default value is "exchange"
    :return: the full URL to the endpoint
    """
    return CONSTANTS.DYDX_V4_REST_URL + path_url


def build_api_factory(
        throttler: AsyncThrottler = None,
) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        rest_pre_processors=[
            DydxV4PerpetualRESTPreProcessor(),
        ],
    )
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(throttler: AsyncThrottler, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> float:
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    rest_assistant = await api_factory.get_rest_assistant()
    url = public_rest_url(CONSTANTS.PATH_TIME)
    limit_id = CONSTANTS.LIMIT_ID_GET
    response = await rest_assistant.execute_request(
        url=url,
        throttler_limit_id=limit_id,
        method=RESTMethod.GET,
    )
    server_time = float(response["epoch"])

    return server_time


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    api_factory = WebAssistantsFactory(throttler=throttler)
    return api_factory


def is_exchange_information_valid(rule: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    if rule["status"] == "ACTIVE":
        valid = True
    else:
        valid = False
    return valid



import asyncio
import time
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.derivative.pacifica_perpetual import (
    pacifica_perpetual_constants as CONSTANTS,
    pacifica_perpetual_web_utils as web_utils,
)
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.pacifica_perpetual.pacifica_perpetual_derivative import (
        PacificaPerpetualDerivative,
    )


class PacificaPerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):
    _logger: Optional[HummingbotLogger] = None

    def __init__(
        self,
        trading_pairs: List[str],
        connector: "PacificaPerpetualDerivative",
        api_factory: WebAssistantsFactory,
        domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._ping_task: Optional[asyncio.Task] = None

    async def get_last_traded_prices(self, trading_pairs: List[str], domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    def _get_headers(self) -> Dict[str, str]:
        headers = {}
        if self._connector.api_config_key:
            headers["PF-API-KEY"] = self._connector.api_config_key
        return headers

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        https://docs.pacifica.fi/api-documentation/api/rest-api/markets/get-orderbook

        {
            "success": true,
            "data": {
                "s": "BTC",
                "l": [
                [
                    {
                    "p": "106504",
                    "a": "0.26203",
                    "n": 1
                    },
                    {
                    "p": "106498",
                    "a": "0.29281",
                    "n": 1
                    }
                ],
                [
                    {
                    "p": "106559",
                    "a": "0.26802",
                    "n": 1
                    },
                    {
                    "p": "106564",
                    "a": "0.3002",
                    "n": 1
                    },
                ]
                ],
                "t": 1751370536325
            },
            "error": null,
            "code": null
        }
        """
        rest_assistant = await self._api_factory.get_rest_assistant()
        params = {"symbol": await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)}

        response = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.GET_MARKET_ORDER_BOOK_SNAPSHOT_PATH_URL, domain=self._domain),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.GET_MARKET_ORDER_BOOK_SNAPSHOT_PATH_URL,
            headers=self._get_headers()
        )

        if not response.get("success") is True:
            raise ValueError(f"[get_order_book_snapshot] Failed to get order book snapshot for {trading_pair}: {response}")

        if not response.get("data", []):
            raise ValueError(f"[get_order_book_snapshot] No data when requesting order book snapshot for {trading_pair}: {response}")

        return response["data"]

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        order_book_snapshot_data = await self._request_order_book_snapshot(trading_pair)
        order_book_snapshot_timestamp = order_book_snapshot_data["t"] / 1000

        return OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": trading_pair,
            "update_id": order_book_snapshot_timestamp,
            "bids": [(bids["p"], bids["a"]) for bids in order_book_snapshot_data["l"][0]],
            "asks": [(asks["p"], asks["a"]) for asks in order_book_snapshot_data["l"][1]]
        }, timestamp=order_book_snapshot_timestamp)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        """
        https://docs.pacifica.fi/api-documentation/api/rest-api/markets/get-prices

        {
            "success": true,
            "data": [
                {
                "funding": "0.00010529",
                "mark": "1.084819",
                "mid": "1.08615",
                "next_funding": "0.00011096",
                "open_interest": "3634796",
                "oracle": "1.084524",
                "symbol": "XPL",
                "timestamp": 1759222967974,
                "volume_24h": "20896698.0672",
                "yesterday_price": "1.3412"
                }
            ],
            "error": null,
            "code": null
        }

        Index price = Oracle price
        Next funding timestamp = :00 of next hour
        """
        rest_assistant = await self._api_factory.get_rest_assistant()
        symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

        response = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.GET_PRICES_PATH_URL, domain=self._domain),
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.GET_PRICES_PATH_URL,
            headers=self._get_headers()
        )

        if not response.get("success") is True:
            raise ValueError(f"[get_funding_info] Failed to get price info for {trading_pair}: {response}")

        if not response.get("data", []):
            raise ValueError(f"[get_funding_info] No data when requesting price info for {trading_pair}: {response}")

        for price_info in response["data"]:
            if price_info["symbol"] == symbol:
                break
        else:
            raise ValueError(f"[get_funding_info] Failed to get price info for {trading_pair}: {response}")

        return FundingInfo(
            trading_pair=trading_pair,
            index_price=Decimal(price_info["oracle"]),
            mark_price=Decimal(price_info["mark"]),
            next_funding_utc_timestamp=int((time.time() // 3600 + 1) * 3600),
            rate=Decimal(price_info["funding"]),
        )

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()

        await ws.connect(ws_url=web_utils.wss_url(self._domain), ws_headers=self._get_headers())
        self._ping_task = safe_ensure_future(self._ping_loop(ws))
        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        try:
            # OB snapshots
            for trading_pair in self._trading_pairs:
                payload = {
                    "method": "subscribe",
                    "params": {
                        "source": "book",
                        "symbol": await self._connector.exchange_symbol_associated_to_pair(trading_pair),
                        "agg_level": 1,
                    },
                }
                subscribe_request = WSJSONRequest(payload=payload)
                await ws.send(subscribe_request)

            # no OB diffs

            # trades
            for trading_pair in self._trading_pairs:
                payload = {
                    "method": "subscribe",
                    "params": {
                        "source": "trades",
                        "symbol": await self._connector.exchange_symbol_associated_to_pair(trading_pair),
                    },
                }
                subscribe_request = WSJSONRequest(payload=payload)
                await ws.send(subscribe_request)

            # funding info

            payload = {
                "method": "subscribe",
                "params": {
                    "source": "prices",
                },
            }
            subscribe_request = WSJSONRequest(payload=payload)
            await ws.send(subscribe_request)

            self.logger().info("Subscribed to public order book and trade channels...")

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading pairs.")
            raise

    async def _on_order_stream_interruption(self, websocket_assistant: Optional[WSAssistant] = None):
        await super()._on_order_stream_interruption(websocket_assistant)
        if self._ping_task is not None:
            self._ping_task.cancel()
            self._ping_task = None

    async def _ping_loop(self, ws: WSAssistant):
        while True:
            try:
                await asyncio.sleep(CONSTANTS.WS_PING_INTERVAL)
                ping_request = WSJSONRequest(payload={"method": "ping"})
                await ws.send(ping_request)
            except asyncio.CancelledError:
                raise
            except RuntimeError as e:
                if "WS is not connected" in str(e):
                    return
                raise
            except Exception:
                self.logger().warning("Error sending ping to Pacifica WebSocket", exc_info=True)
                await asyncio.sleep(5.0)  # Wait before retrying

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        https://docs.pacifica.fi/api-documentation/api/websocket/subscriptions/orderbook

        {
            "channel": "book",
            "data": {
                "l": [
                [
                    {
                    "a": "37.86",
                    "n": 4,
                    "p": "157.47"
                    },
                    // ... other aggegated bid levels
                ],
                [
                    {
                    "a": "12.7",
                    "n": 2,
                    "p": "157.49"
                    },
                    {
                    "a": "44.45",
                    "n": 3,
                    "p": "157.5"
                    },
                    // ... other aggregated ask levels
                ]
                ],
                "s": "SOL",
                "t": 1749051881187,
                "li": 1559885104
            }
        }
        """
        snapshot_data = raw_message["data"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=snapshot_data["s"])
        snapshot_timestamp = snapshot_data["t"] / 1000  # exchange provides time in ms
        update_id = snapshot_data["li"]

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": [(bid["p"], bid["a"]) for bid in snapshot_data["l"][0]],
            "asks": [(ask["p"], ask["a"]) for ask in snapshot_data["l"][1]],
        }
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            order_book_message_content,
            snapshot_timestamp)

        message_queue.put_nowait(snapshot_msg)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        https://docs.pacifica.fi/api-documentation/api/websocket/subscriptions/trades

        {
            "channel": "trades",
            "data": [
                {
                "u": "42trU9A5...",
                "h": 80062522,
                "s": "BTC",
                "a": "0.00001",
                "p": "89471",
                "d": "close_short",
                "tc": "normal",
                "t": 1765018379085,
                "li": 1559885104
                }
            ]
        }

        Trade side:
        (*) open_long
        (*) open_short
        (*) close_long
        (*) close_short
        """
        trade_updates = raw_message["data"]

        for trade_data in trade_updates:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=trade_data["s"])
            message_content = {
                "trade_id": trade_data["h"],  # we use history id as trade id
                "update_id": trade_data["li"],
                "trading_pair": trading_pair,
                "trade_type": float(TradeType.BUY.value) if trade_data["d"] in ("open_long", "close_short") else float(TradeType.SELL.value),
                "amount": trade_data["a"],
                "price": trade_data["p"]
            }
            trade_message: Optional[OrderBookMessage] = OrderBookMessage(
                message_type=OrderBookMessageType.TRADE,
                content=message_content,
                timestamp=trade_data["t"] / 1000  # originally it's time in ms
            )

            message_queue.put_nowait(trade_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        https://docs.pacifica.fi/api-documentation/api/websocket/subscriptions/prices

        {
            "channel": "prices",
            "data": [
                {
                    "funding": "0.0000125",
                    "mark": "105473",
                    "mid": "105476",
                    "next_funding": "0.0000125",
                    "open_interest": "0.00524",
                    "oracle": "105473",
                    "symbol": "BTC",
                    "timestamp": 1749051612681,
                    "volume_24h": "63265.87522",
                    "yesterday_price": "955476"
                }
                // ... other symbol prices
            ],
        }

        Index price = Oracle price
        Next funding timestamp = :00 of next hour
        """
        for price_entry in raw_message["data"]:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(price_entry["symbol"])
            if trading_pair not in self._trading_pairs:
                continue

            info_update = FundingInfoUpdate(
                trading_pair=trading_pair,
                index_price=Decimal(price_entry["oracle"]),
                mark_price=Decimal(price_entry["mark"]),
                next_funding_utc_timestamp=int((time.time() // 3600 + 1) * 3600),
                rate=Decimal(price_entry["funding"])
            )

            message_queue.put_nowait(info_update)

            self._connector.set_pacifica_price(
                trading_pair,
                timestamp=price_entry["timestamp"] / 1000,
                index_price=Decimal(price_entry["oracle"]),
                mark_price=Decimal(price_entry["mark"]),
            )

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "data" in event_message:
            event_channel = event_message["channel"]
            if event_channel == CONSTANTS.WS_ORDER_BOOK_SNAPSHOT_CHANNEL:
                channel = self._snapshot_messages_queue_key
            elif event_channel == CONSTANTS.WS_TRADES_CHANNEL:
                channel = self._trade_messages_queue_key
            elif event_channel == CONSTANTS.WS_PRICES_CHANNEL:
                channel = self._funding_info_messages_queue_key
        return channel

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribes to order book and trade channels for a single trading pair
        on the existing WebSocket connection.

        :param trading_pair: the trading pair to subscribe to
        :return: True if subscription was successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot subscribe to {trading_pair}: WebSocket not connected"
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            # Subscribe to order book snapshots
            book_payload = {
                "method": "subscribe",
                "params": {
                    "source": "book",
                    "symbol": symbol,
                    "agg_level": 1,
                },
            }
            subscribe_book_request = WSJSONRequest(payload=book_payload)

            # Subscribe to trades
            trades_payload = {
                "method": "subscribe",
                "params": {
                    "source": "trades",
                    "symbol": symbol,
                },
            }
            subscribe_trades_request = WSJSONRequest(payload=trades_payload)

            await self._ws_assistant.send(subscribe_book_request)
            await self._ws_assistant.send(subscribe_trades_request)

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to {trading_pair} order book and trade channels")
            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error subscribing to {trading_pair}")
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribes from order book and trade channels for a single trading pair
        on the existing WebSocket connection.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if unsubscription was successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket not connected"
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            # Unsubscribe from order book snapshots
            book_payload = {
                "method": "unsubscribe",
                "params": {
                    "source": "book",
                    "symbol": symbol,
                    "agg_level": 1,
                },
            }
            unsubscribe_book_request = WSJSONRequest(payload=book_payload)

            # Unsubscribe from trades
            trades_payload = {
                "method": "unsubscribe",
                "params": {
                    "source": "trades",
                    "symbol": symbol,
                },
            }
            unsubscribe_trades_request = WSJSONRequest(payload=trades_payload)

            await self._ws_assistant.send(unsubscribe_book_request)
            await self._ws_assistant.send(unsubscribe_trades_request)

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from {trading_pair} order book and trade channels")

            # TODO (dizpers): to be 100% sure we should actually wait until the copy of unsub message is received

            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error unsubscribing from {trading_pair}")
            return False



import asyncio
import json
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.bitmart import bitmart_constants as CONSTANTS, bitmart_utils as utils
from hummingbot.connector.exchange.bitmart.bitmart_auth import BitmartAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest, WSResponse
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.bitmart.bitmart_exchange import BitmartExchange


class BitmartAPIUserStreamDataSource(UserStreamTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None

    def __init__(
        self,
        auth: BitmartAuth,
        trading_pairs: List[str],
        connector: 'BitmartExchange',
        api_factory: WebAssistantsFactory
    ):
        super().__init__()
        self._auth: BitmartAuth = auth
        self._trading_pairs = trading_pairs
        self._connector = connector
        self._api_factory = api_factory

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange
        """

        ws: WSAssistant = await self._get_ws_assistant()
        await ws.connect(
            ws_url=CONSTANTS.WSS_PRIVATE_URL,
            ping_timeout=CONSTANTS.WS_PING_TIMEOUT)

        payload = {
            "op": "login",
            "args": self._auth.websocket_login_parameters()
        }

        login_request: WSJSONRequest = WSJSONRequest(payload=payload)

        async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_SUBSCRIBE):
            await ws.send(login_request)

        response: WSResponse = await ws.receive()
        message = response.data
        if "errorCode" in message or "error_code" in message or message.get("event") != "login":
            self.logger().error("Error authenticating the private websocket connection")
            raise IOError(f"Private websocket connection authentication failed ({message})")

        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        try:
            symbols = [await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                       for trading_pair in self._trading_pairs]

            payload = {
                "op": "subscribe",
                "args": [f"{CONSTANTS.PRIVATE_ORDER_PROGRESS_CHANNEL_NAME}:{symbol}" for symbol in symbols]
            }
            subscribe_request: WSJSONRequest = WSJSONRequest(payload=payload)

            async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_SUBSCRIBE):
                await websocket_assistant.send(subscribe_request)
            self.logger().info("Subscribed to private account and orders channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        async for ws_response in websocket_assistant.iter_messages():
            data: Dict[str, Any] = ws_response.data
            decompressed_data = utils.decompress_ws_message(data)
            try:
                if isinstance(decompressed_data, str):
                    json_data = json.loads(decompressed_data)
                else:
                    json_data = decompressed_data
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().warning(f"Invalid event message received through the order book data source "
                                      f"connection ({decompressed_data})")
                continue

            if "errorCode" in json_data or "errorMessage" in json_data:
                raise ValueError(f"Error message received in the order book data source: {json_data}")

            await self._process_event_message(event_message=json_data, queue=queue)

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if len(event_message) > 0 and "table" in event_message and "data" in event_message:
            queue.put_nowait(event_message)

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant



"""
hummingbot.client.config.config_var defines ConfigVar. One of its parameters is a validator, a function that takes a
string and determines whether it is valid input. This file contains many validator functions that are used by various
hummingbot ConfigVars.
"""

import re
import time
from datetime import datetime
from decimal import Decimal
from typing import Optional


def validate_exchange(value: str) -> Optional[str]:
    """
    Restrict valid connectors to spot connectors
    """
    from hummingbot.client.settings import AllConnectorSettings
    if value not in AllConnectorSettings.get_exchange_names():
        return f"Invalid exchange, please choose value from {AllConnectorSettings.get_exchange_names()}"


def validate_derivative(value: str) -> Optional[str]:
    """
    Restrict valid connectors to perpetual connectors
    """
    from hummingbot.client.settings import AllConnectorSettings
    if value not in AllConnectorSettings.get_derivative_names():
        return f"Invalid derivative, please choose value from {AllConnectorSettings.get_derivative_names()}"


def validate_connector(value: str) -> Optional[str]:
    """
    Restrict valid connectors to ALL spot connectors, including paper trade and Gateway
    """
    from hummingbot.client.settings import GATEWAY_CONNECTORS, AllConnectorSettings
    valid_connectors = set(AllConnectorSettings.get_connector_settings().keys())
    valid_connectors.update(AllConnectorSettings.paper_trade_connectors_names)
    valid_connectors.update(GATEWAY_CONNECTORS)

    if value not in valid_connectors:
        all_options = sorted(valid_connectors)
        return f"Invalid connector, please choose value from {all_options}"


def validate_strategy(value: str) -> Optional[str]:
    """
    Restrict valid derivatives to the strategy file names
    """
    from hummingbot.client.settings import STRATEGIES
    if value not in STRATEGIES:
        return f"Invalid strategy, please choose value from {STRATEGIES}"


def validate_decimal(value: str, min_value: Decimal = None, max_value: Decimal = None, inclusive=True) -> Optional[str]:
    """
    Parse a decimal value from a string. This value can also be clamped.
    """
    try:
        decimal_value = Decimal(value)
    except Exception:
        return f"{value} is not in decimal format."
    if inclusive:
        if min_value is not None and max_value is not None:
            if not (Decimal(str(min_value)) <= decimal_value <= Decimal(str(max_value))):
                return f"Value must be between {min_value} and {max_value}."
        elif min_value is not None and not decimal_value >= Decimal(str(min_value)):
            return f"Value cannot be less than {min_value}."
        elif max_value is not None and not decimal_value <= Decimal(str(max_value)):
            return f"Value cannot be more than {max_value}."
    else:
        if min_value is not None and max_value is not None:
            if not (Decimal(str(min_value)) < decimal_value < Decimal(str(max_value))):
                return f"Value must be between {min_value} and {max_value} (exclusive)."
        elif min_value is not None and not decimal_value > Decimal(str(min_value)):
            return f"Value must be more than {min_value}."
        elif max_value is not None and not decimal_value < Decimal(str(max_value)):
            return f"Value must be less than {max_value}."


def validate_market_trading_pair(market: str, value: str) -> Optional[str]:
    """
    Since trading pair validation and autocomplete are UI optimizations that do not impact bot performances,
    in case of network issues or slow wifi, this check returns true and does not prevent users from proceeding,
    """
    from hummingbot.core.utils.trading_pair_fetcher import TradingPairFetcher
    trading_pair_fetcher: TradingPairFetcher = TradingPairFetcher.get_instance()
    if trading_pair_fetcher.ready:
        trading_pairs = trading_pair_fetcher.trading_pairs.get(market, [])
        if len(trading_pairs) == 0:
            return None
        elif value not in trading_pairs:
            return f"{value} is not an active market on {market}."


def validate_bool(value: str) -> Optional[str]:
    """
    Permissively interpret a string as a boolean
    """
    valid_values = ('true', 'yes', 'y', 'false', 'no', 'n')
    if value.lower() not in valid_values:
        return f"Invalid value, please choose value from {valid_values}"


def validate_int(value: str, min_value: int = None, max_value: int = None, inclusive=True) -> Optional[str]:
    """
    Parse an int value from a string. This value can also be clamped.
    """
    try:
        int_value = int(value)
    except Exception:
        return f"{value} is not in integer format."
    if inclusive:
        if min_value is not None and max_value is not None:
            if not (min_value <= int_value <= max_value):
                return f"Value must be between {min_value} and {max_value}."
        elif min_value is not None and not int_value >= min_value:
            return f"Value cannot be less than {min_value}."
        elif max_value is not None and not int_value <= max_value:
            return f"Value cannot be more than {max_value}."
    else:
        if min_value is not None and max_value is not None:
            if not (min_value < int_value < max_value):
                return f"Value must be between {min_value} and {max_value} (exclusive)."
        elif min_value is not None and not int_value > min_value:
            return f"Value must be more than {min_value}."
        elif max_value is not None and not int_value < max_value:
            return f"Value must be less than {max_value}."


def validate_float(value: str, min_value: float = None, max_value: float = None, inclusive=True) -> Optional[str]:
    """
    Parse an float value from a string. This value can also be clamped.
    """
    try:
        float_value = float(value)
    except Exception:
        return f"{value} is not in integer format."
    if inclusive:
        if min_value is not None and max_value is not None:
            if not (min_value <= float_value <= max_value):
                return f"Value must be between {min_value} and {max_value}."
        elif min_value is not None and not float_value >= min_value:
            return f"Value cannot be less than {min_value}."
        elif max_value is not None and not float_value <= max_value:
            return f"Value cannot be more than {max_value}."
    else:
        if min_value is not None and max_value is not None:
            if not (min_value < float_value < max_value):
                return f"Value must be between {min_value} and {max_value} (exclusive)."
        elif min_value is not None and not float_value > min_value:
            return f"Value must be more than {min_value}."
        elif max_value is not None and not float_value < max_value:
            return f"Value must be less than {max_value}."


def validate_datetime_iso_string(value: str) -> Optional[str]:
    try:
        datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
    except ValueError:
        return "Incorrect date time format (expected is YYYY-MM-DD HH:MM:SS)"


def validate_time_iso_string(value: str) -> Optional[str]:
    try:
        time.strptime(value, '%H:%M:%S')
    except ValueError:
        return "Incorrect time format (expected is HH:MM:SS)"


def validate_with_regex(value: str, pattern: str, error_message: str) -> Optional[str]:
    """
    Validate a string using a regex pattern.
    """
    if not re.match(pattern, value):
        return error_message



import logging
from typing import Any, Awaitable, Callable, Dict, List, Optional

from hummingbot.client.config.config_helpers import ClientConfigAdapter
from hummingbot.client.settings import AllConnectorSettings, ConnectorSetting
from hummingbot.logger import HummingbotLogger

from ...client.config.security import Security
from .async_utils import safe_ensure_future


class TradingPairFetcher:
    _sf_shared_instance: "TradingPairFetcher" = None
    _tpf_logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._tpf_logger is None:
            cls._tpf_logger = logging.getLogger(__name__)
        return cls._tpf_logger

    @classmethod
    def get_instance(cls, client_config_map: Optional["ClientConfigAdapter"] = None) -> "TradingPairFetcher":
        if cls._sf_shared_instance is None:
            client_config_map = client_config_map or cls._get_client_config_map()
            cls._sf_shared_instance = TradingPairFetcher(client_config_map)
        return cls._sf_shared_instance

    def __init__(self, client_config_map: ClientConfigAdapter):
        self.ready = False
        self.trading_pairs: Dict[str, Any] = {}
        self.fetch_pairs_from_all_exchanges = client_config_map.hb_config.fetch_pairs_from_all_exchanges
        self._fetch_task = safe_ensure_future(self.fetch_all(client_config_map))

    def _fetch_pairs_from_connector_setting(
            self,
            connector_setting: ConnectorSetting,
            connector_name: Optional[str] = None):
        connector_name = connector_name or connector_setting.name
        connector = connector_setting.non_trading_connector_instance_with_default_configuration()
        safe_ensure_future(self.call_fetch_pairs(connector.all_trading_pairs(), connector_name))

    async def fetch_all(self, client_config_map: ClientConfigAdapter):
        await Security.wait_til_decryption_done()
        connector_settings = self._all_connector_settings()
        for conn_setting in connector_settings.values():
            # XXX(martin_kou): Some connectors, e.g. uniswap v3, aren't completed yet. Ignore if you can't find the
            # data source module for them.
            try:
                if conn_setting.base_name().endswith("paper_trade"):
                    self._fetch_pairs_from_connector_setting(
                        connector_setting=connector_settings[conn_setting.parent_name],
                        connector_name=conn_setting.name
                    )
                elif not self.fetch_pairs_from_all_exchanges:
                    if conn_setting.connector_connected():
                        self._fetch_pairs_from_connector_setting(connector_setting=conn_setting)
                else:
                    self._fetch_pairs_from_connector_setting(connector_setting=conn_setting)
            except ModuleNotFoundError:
                continue
            except Exception:
                self.logger().exception(f"An error occurred when fetching trading pairs for {conn_setting.name}."
                                        "Please check the logs")
        self.ready = True

    async def call_fetch_pairs(self, fetch_fn: Callable[[], Awaitable[List[str]]], exchange_name: str):
        try:
            pairs = await fetch_fn
            self.trading_pairs[exchange_name] = pairs
        except Exception:
            self.logger().error(f"Connector {exchange_name} failed to retrieve its trading pairs. "
                                f"Trading pairs autocompletion won't work.", exc_info=True)
            # In case of error just assign empty list, this is st. the bot won't stop working
            self.trading_pairs[exchange_name] = []

    def _all_connector_settings(self) -> Dict[str, ConnectorSetting]:
        # Method created to enabling patching in unit tests
        return AllConnectorSettings.get_connector_settings()

    @staticmethod
    def _get_client_config_map() -> "ClientConfigAdapter":
        from hummingbot.client.hummingbot_application import HummingbotApplication

        client_config_map = HummingbotApplication.main_application().client_config_map
        return client_config_map



import asyncio
import functools
from asyncio import Task
from decimal import Decimal
from typing import Dict, List, Optional, Union

from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils import async_ttl_cache
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.data_feed.coin_gecko_data_feed import CoinGeckoDataFeed
from hummingbot.data_feed.coin_gecko_data_feed.coin_gecko_constants import COOLOFF_AFTER_BAN, CoinGeckoAPITier


class CoinGeckoRateSource(RateSourceBase):
    def __init__(
        self,
        extra_token_ids: List[str],
        api_key: str = "",
        api_tier: CoinGeckoAPITier = CoinGeckoAPITier.PUBLIC,
    ):
        super().__init__()
        self._coin_gecko_supported_vs_tokens: Optional[List[str]] = None
        self._coin_gecko_data_feed: Optional[CoinGeckoDataFeed] = None  # delayed because of circular reference
        self._extra_token_ids = extra_token_ids
        self._api_key = api_key
        self._api_tier = api_tier
        self._rate_limit_exceeded = asyncio.Event()
        self._lock = asyncio.Lock()

    @property
    def name(self) -> str:
        return "coin_gecko"

    @property
    def extra_token_ids(self) -> List[str]:
        return self._extra_token_ids

    @extra_token_ids.setter
    def extra_token_ids(self, new_ids: List[str]):
        self._extra_token_ids = new_ids

    @property
    def api_key(self) -> str:
        return self._api_key

    @api_key.setter
    def api_key(self, new_api_key: str):
        self._api_key = new_api_key
        # Update data feed if it already exists
        if self._coin_gecko_data_feed is not None:
            self._coin_gecko_data_feed._api_key = new_api_key
            # Update rate limits directly from the tier
            self._coin_gecko_data_feed._api_factory._throttler._rate_limits = self._api_tier.value.rate_limits

    @property
    def api_tier(self) -> CoinGeckoAPITier:
        return self._api_tier

    @api_tier.setter
    def api_tier(self, new_tier: CoinGeckoAPITier):
        self._api_tier = new_tier
        # Update data feed if it already exists
        if self._coin_gecko_data_feed is not None:
            self._coin_gecko_data_feed._api_tier = new_tier
            # Update rate limits directly from the tier
            self._coin_gecko_data_feed._api_factory._throttler._rate_limits = new_tier.value.rate_limits

    def try_event(self, fn):
        @functools.wraps(fn)
        async def try_raise_event(*args, **kwargs):
            while True:
                # If the rate limit has been exceeded, wait for the cool-off period to pass
                if self._rate_limit_exceeded.is_set():
                    await self._rate_limit_exceeded.wait()

                try:
                    res = await fn(*args, **kwargs)
                    return res
                except IOError as e:
                    # This is from exceeding the server's rate limit, signal the issue and wait for post-ban cool-off
                    self.logger().warning("Rate limit exceeded with:")
                    self.logger().warning(f"   {e}")
                    self.logger().warning("   Report to development team")
                    self._rate_limit_exceeded.set()
                    # This is the cool-off after a ban
                    await self._sleep(COOLOFF_AFTER_BAN)
                    self.logger().info(f"   Continuing after {COOLOFF_AFTER_BAN} seconds")
                    self._rate_limit_exceeded.clear()
                except Exception as e:
                    self.logger().error(f"Unhandled error in CoinGecko rate source response: {str(e)}", exc_info=True)
                    raise Exception(f"Unhandled error in CoinGecko rate source response: {str(e)}")

        return try_raise_event

    @async_ttl_cache(ttl=COOLOFF_AFTER_BAN, maxsize=1)
    async def get_prices(self, quote_token: Optional[str] = None) -> Dict[str, Decimal]:
        """
        Fetches the first 2500 CoinGecko prices ordered by market cap to ~ 500K USD

        :param quote_token: The quote token for which to fetch prices
        :return A dictionary of trading pairs and prices
        """
        await self._lock.acquire()

        if quote_token is None:
            raise NotImplementedError("Must supply a quote token to fetch prices for CoinGecko")
        self._ensure_data_feed()
        vs_currency = quote_token.lower()
        results = {}
        if not self._coin_gecko_supported_vs_tokens:
            self._coin_gecko_supported_vs_tokens = await self.try_event(
                self._coin_gecko_data_feed.get_supported_vs_tokens)()

        if vs_currency not in self._coin_gecko_supported_vs_tokens:
            vs_currency = "usd"

        # Extra tokens
        r = await self.try_event(self._get_coin_gecko_extra_token_prices)(vs_currency)
        results.update(r)

        # Coin Gecko returns 250 assets max per page, 2500th is around 500K USD market cap (as of 2/2023)
        tasks: List[Task] = []
        for page_no in range(1, 8):
            tasks.append(asyncio.create_task(self._get_coin_gecko_prices_by_page(vs_currency, page_no, None)))

        try:
            task_results = await self.try_event(safe_gather)(*tasks, return_exceptions=False)
        except Exception:
            self.logger().error(
                "Unexpected error while retrieving rates from Coingecko. Check the log file for more info.")
            raise

        # Collect the results
        for i, task_result in enumerate(task_results):
            results.update(task_result)

        self._lock.release()
        return results

    def _ensure_data_feed(self):
        if self._coin_gecko_data_feed is None:
            self._coin_gecko_data_feed = CoinGeckoDataFeed(
                api_key=self._api_key,
                api_tier=self._api_tier,
            )

    async def _get_coin_gecko_prices_by_page(self,
                                             vs_currency: str,
                                             page_no: int,
                                             category: Union[str, None]) -> Dict[str, Decimal]:
        """
        Fetches CoinGecko prices by page number.

        :param vs_currency: A currency (crypto or fiat) to get prices of tokens in, see
        https://api.coingecko.com/api/v3/simple/supported_vs_currencies for the current supported list
        :param page_no: The page number
        :param category | None: category to filter tokens to get from the provider (specifying one limits to 50 results)

        :return: A dictionary of trading pairs and prices (50 results max if a category is provided)
        """
        results = {}
        resp = await self.try_event(self._coin_gecko_data_feed.get_prices_by_page)(vs_currency=vs_currency,
                                                                                   page_no=page_no, category=category)

        for record in resp:
            pair = combine_to_hb_trading_pair(base=record['symbol'].upper(), quote=vs_currency.upper())
            if record["current_price"]:
                results[pair] = Decimal(str(record["current_price"]))
        return results

    async def _get_coin_gecko_extra_token_prices(self, vs_currency: str) -> Dict[str, Decimal]:
        """
        Fetches CoinGecko prices for the configured extra tokens.

        :param vs_currency: A currency (crypto or fiat) to get prices of tokens in, see
        https://api.coingecko.com/api/v3/simple/supported_vs_currencies for the current supported list

        :return: A dictionary of trading pairs and prices
        """
        results = {}
        # TODO: Should we force hummingbot to be included?
        # self._extra_token_ids.append("hummingbot") - This fails the tests, not sure why
        if self._extra_token_ids:
            resp = await self.try_event(self._coin_gecko_data_feed.get_prices_by_token_id)(vs_currency=vs_currency,
                                                                                           token_ids=self._extra_token_ids)
            for record in resp:
                pair = combine_to_hb_trading_pair(base=record["symbol"].upper(), quote=vs_currency.upper())
                if record["current_price"]:
                    results[pair] = Decimal(str(record["current_price"]))
        return results

    async def _sleep(self, delay: float):
        await asyncio.sleep(delay)



from decimal import Decimal
from functools import lru_cache
from typing import Dict, List, Optional, Tuple, Union

from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.core.data_type.common import OrderType, PositionAction, PriceType, TradeType
from hummingbot.core.data_type.order_candidate import OrderCandidate
from hummingbot.core.event.event_forwarder import SourceInfoEventForwarder
from hummingbot.core.event.events import (
    BuyOrderCompletedEvent,
    BuyOrderCreatedEvent,
    MarketEvent,
    MarketOrderFailureEvent,
    OrderCancelledEvent,
    OrderFilledEvent,
    SellOrderCompletedEvent,
    SellOrderCreatedEvent,
)
from hummingbot.strategy.strategy_v2_base import StrategyV2Base
from hummingbot.strategy_v2.executors.data_types import ExecutorConfigBase
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executors import CloseType
from hummingbot.strategy_v2.models.executors_info import ExecutorInfo
from hummingbot.strategy_v2.runnable_base import RunnableBase


class ExecutorBase(RunnableBase):
    """
    Base class for all executors. Executors are responsible for executing orders based on the strategy.
    """

    def __init__(self, strategy: StrategyV2Base, connectors: List[str], config: ExecutorConfigBase, update_interval: float = 0.5):
        """
        Initializes the executor with the given strategy, connectors and update interval.

        :param strategy: The strategy to be used by the executor.
        :param connectors: The connectors to be used by the executor.
        :param update_interval: The update interval for the executor.
        """
        super().__init__(update_interval)
        self.config = config
        self.close_type: Optional[CloseType] = None
        self.close_timestamp: Optional[float] = None
        self._strategy: StrategyV2Base = strategy
        self._held_position_orders = []  # Keep track of orders that become held positions
        self.connectors = {connector_name: connector for connector_name, connector in strategy.connectors.items() if
                           connector_name in connectors}

        # Event forwarders for different order events
        self._create_buy_order_forwarder = SourceInfoEventForwarder(self.process_order_created_event)
        self._create_sell_order_forwarder = SourceInfoEventForwarder(self.process_order_created_event)
        self._fill_order_forwarder = SourceInfoEventForwarder(self.process_order_filled_event)
        self._complete_buy_order_forwarder = SourceInfoEventForwarder(self.process_order_completed_event)
        self._complete_sell_order_forwarder = SourceInfoEventForwarder(self.process_order_completed_event)
        self._cancel_order_forwarder = SourceInfoEventForwarder(self.process_order_canceled_event)
        self._failed_order_forwarder = SourceInfoEventForwarder(self.process_order_failed_event)

        # Pairs of market events and their corresponding event forwarders
        self._event_pairs: List[Tuple[MarketEvent, SourceInfoEventForwarder]] = [
            (MarketEvent.OrderCancelled, self._cancel_order_forwarder),
            (MarketEvent.BuyOrderCreated, self._create_buy_order_forwarder),
            (MarketEvent.SellOrderCreated, self._create_sell_order_forwarder),
            (MarketEvent.OrderFilled, self._fill_order_forwarder),
            (MarketEvent.BuyOrderCompleted, self._complete_buy_order_forwarder),
            (MarketEvent.SellOrderCompleted, self._complete_sell_order_forwarder),
            (MarketEvent.OrderFailure, self._failed_order_forwarder),
        ]

    @property
    def status(self):
        """
        Returns the status of the executor.
        """
        return self._status

    @property
    def is_trading(self):
        """
        Returns whether the executor is trading.
        """
        return self.is_active and self.net_pnl_quote != 0

    @property
    def filled_amount_quote(self):
        """
        Returns the filled amount in quote currency.
        """
        return Decimal("0")

    @property
    def is_active(self):
        """
        Returns whether the executor is open or trading.
        """
        return self._status in [RunnableStatus.RUNNING, RunnableStatus.NOT_STARTED]

    @property
    def is_closed(self):
        """
        Returns whether the executor is closed.
        """
        return self._status == RunnableStatus.TERMINATED

    @property
    def executor_info(self) -> ExecutorInfo:
        """
        Returns the executor info.
        """
        def _safe_decimal(value) -> Decimal:
            d = Decimal(str(value))
            return d if d.is_finite() else Decimal("0")

        ei = ExecutorInfo(
            id=self.config.id,
            timestamp=self.config.timestamp,
            type=self.config.type,
            status=self.status,
            close_type=self.close_type,
            close_timestamp=self.close_timestamp,
            config=self.config,
            net_pnl_pct=_safe_decimal(self.net_pnl_pct),
            net_pnl_quote=_safe_decimal(self.net_pnl_quote),
            cum_fees_quote=_safe_decimal(self.cum_fees_quote),
            filled_amount_quote=_safe_decimal(self.filled_amount_quote),
            is_active=self.is_active,
            is_trading=self.is_trading,
            custom_info=self.get_custom_info(),
            controller_id=self.config.controller_id,
        )
        return ei

    def get_custom_info(self) -> Dict:
        """
        Returns the custom info of the executor. Returns an empty dictionary by default, and can be reimplemented
        by subclasses.
        """
        return {}

    @staticmethod
    def is_perpetual_connector(connector_name: str):
        """
        Returns whether the specified connector is a perpetual connector.

        :param connector_name: The name of the connector.
        :return: True if the connector is a perpetual connector, False otherwise.
        """
        return "perpetual" in connector_name.lower()

    @staticmethod
    @lru_cache(maxsize=10)
    def is_amm_connector(exchange: str) -> bool:
        return exchange in sorted(
            AllConnectorSettings.get_gateway_amm_connector_names()
        )

    def start(self):
        """
        Starts the executor and registers the events.
        """
        super().start()
        self.register_events()

    def stop(self):
        """
        Stops the executor and unregisters the events.
        """
        self.close_timestamp = self._strategy.current_timestamp
        super().stop()
        self.unregister_events()

    async def on_start(self):
        """
        Called when the executor is started.
        """
        await self.validate_sufficient_balance()

    def on_stop(self):
        """
        Called when the executor is stopped.
        """
        pass

    def early_stop(self, keep_position: bool = False):
        """
        This method allows strategy to stop the executor early.
        """
        raise NotImplementedError

    async def validate_sufficient_balance(self):
        """
        Validates that the executor has sufficient balance to place orders.
        """
        raise NotImplementedError

    @property
    def net_pnl_quote(self) -> Decimal:
        """
        Returns the net profit or loss in quote currency.
        """
        return self.get_net_pnl_quote()

    @property
    def net_pnl_pct(self) -> Decimal:
        """
        Returns the net profit or loss in percentage.
        """
        return self.get_net_pnl_pct()

    @property
    def cum_fees_quote(self) -> Decimal:
        """
        Returns the cumulative fees in quote currency.
        """
        return self.get_cum_fees_quote()

    def get_net_pnl_quote(self) -> Decimal:
        """
        Returns the net profit or loss in quote currency.
        """
        raise NotImplementedError

    def get_net_pnl_pct(self) -> Decimal:
        """
        Returns the net profit or loss in percentage.
        """
        raise NotImplementedError

    def get_cum_fees_quote(self) -> Decimal:
        """
        Returns the cumulative fees in quote currency.
        """
        raise NotImplementedError

    def get_in_flight_order(self, connector_name: str, order_id: str):
        """
        Retrieves an in-flight order from the specified connector using the order ID.

        :param connector_name: The name of the connector.
        :param order_id: The ID of the order.
        :return: The in-flight order.
        """
        return self.connectors[connector_name]._order_tracker.fetch_order(client_order_id=order_id)

    def register_events(self):
        """
        Registers the events with the connectors.
        """
        for connector in self.connectors.values():
            for event_pair in self._event_pairs:
                connector.add_listener(event_pair[0], event_pair[1])

    def unregister_events(self):
        """
        Unregisters the events from the connectors.
        """
        for connector in self.connectors.values():
            for event_pair in self._event_pairs:
                connector.remove_listener(event_pair[0], event_pair[1])

    def adjust_order_candidates(self, exchange: str, order_candidates: List[OrderCandidate]) -> List[OrderCandidate]:
        """
        Adjusts the order candidates based on the budget checker of the specified exchange.
        """
        return self.connectors[exchange].budget_checker.adjust_candidates(order_candidates)

    def place_order(self,
                    connector_name: str,
                    trading_pair: str,
                    order_type: OrderType,
                    side: TradeType,
                    amount: Decimal,
                    position_action: PositionAction = PositionAction.NIL,
                    price=Decimal("NaN"),
                    ):
        """
        Places an order with the specified parameters.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for the order.
        :param order_type: The type of the order.
        :param side: The side of the order (buy or sell).
        :param amount: The amount for the order.
        :param position_action: The position action for the order.
        :param price: The price for the order.
        :return: The result of the order placement.
        """
        if side == TradeType.BUY:
            return self._strategy.buy(connector_name, trading_pair, amount, order_type, price, position_action)
        else:
            return self._strategy.sell(connector_name, trading_pair, amount, order_type, price, position_action)

    def get_price(self, connector_name: str, trading_pair: str, price_type: PriceType = PriceType.MidPrice):
        """
        Retrieves the price for the specified trading pair from the specified connector.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair.
        :param price_type: The type of the price.
        :return: The price.
        """
        return self.connectors[connector_name].get_price_by_type(trading_pair, price_type)

    def get_trading_rules(self, connector_name: str, trading_pair: str) -> TradingRule:
        """
        Retrieves the trading rules for the specified trading pair from the specified connector.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair.
        :return: The trading rules.
        """
        return self.connectors[connector_name].trading_rules[trading_pair]

    def get_order_book(self, connector_name: str, trading_pair: str):
        """
        Retrieves the order book for the specified trading pair from the specified connector.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair.
        :return: The order book.
        """
        return self.connectors[connector_name].get_order_book(connector_name, trading_pair)

    def get_balance(self, connector_name: str, asset: str):
        """
        Retrieves the balance of the specified asset from the specified connector.

        :param connector_name: The name of the connector.
        :param asset: The asset.
        :return: The balance.
        """
        return self.connectors[connector_name].get_balance(asset)

    def get_available_balance(self, connector_name: str, asset: str):
        """
        Retrieves the available balance of the specified asset from the specified connector.

        :param connector_name: The name of the connector.
        :param asset: The asset.
        :return: The available balance.
        """
        return self.connectors[connector_name].get_available_balance(asset)

    def get_active_orders(self, connector_name: str):
        """
        Retrieves the active orders from the specified connector.

        :param connector_name: The name of the connector.
        :return: The active orders.
        """
        return self._strategy.get_active_orders(connector_name)

    def process_order_completed_event(self,
                                      event_tag: int,
                                      market: ConnectorBase,
                                      event: Union[BuyOrderCompletedEvent, SellOrderCompletedEvent]):
        """
        Processes the order completed event. This method should be overridden by subclasses.

        :param event_tag: The event tag.
        :param market: The market where the event occurred.
        :param event: The event.
        """
        pass

    def process_order_created_event(self,
                                    event_tag: int,
                                    market: ConnectorBase,
                                    event: Union[BuyOrderCreatedEvent, SellOrderCreatedEvent]):
        """
        Processes the order created event. This method should be overridden by subclasses.

        :param event_tag: The event tag.
        :param market: The market where the event occurred.
        :param event: The event.
        """
        pass

    def process_order_canceled_event(self,
                                     event_tag: int,
                                     market: ConnectorBase,
                                     event: OrderCancelledEvent):
        """
        Processes the order canceled event. This method should be overridden by subclasses.

        :param event_tag: The event tag.
        :param market: The market where the event occurred.
        :param event: The event.
        """
        pass

    def process_order_filled_event(self,
                                   event_tag: int,
                                   market: ConnectorBase,
                                   event: OrderFilledEvent):
        """
        Processes the order filled event. This method should be overridden by subclasses.

        :param event_tag: The event tag.
        :param market: The market where the event occurred.
        :param event: The event.
        """
        pass

    def process_order_failed_event(self,
                                   event_tag: int,
                                   market: ConnectorBase,
                                   event: MarketOrderFailureEvent):
        """
        Processes the order failed event. This method should be overridden by subclasses.

        :param event_tag: The event tag.
        :param market: The market where the event occurred.
        :param event: The event.
        """
        pass



import asyncio
import logging
from decimal import Decimal
from typing import Dict, Optional, Set

from pydantic import BaseModel

from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient
from hummingbot.core.network_base import NetworkBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class TokenBuySellPrice(BaseModel):
    base: str
    quote: str
    connector: str
    chain: str
    network: str
    order_amount_in_base: Decimal
    buy_price: Decimal
    sell_price: Decimal


class AmmGatewayDataFeed(NetworkBase):
    dex_logger: Optional[HummingbotLogger] = None
    _gateway_client: Optional[GatewayHttpClient] = None

    @classmethod
    def get_gateway_client(cls) -> GatewayHttpClient:
        """Class method for lazy initialization of gateway client to avoid duplicate initialization during import"""
        if cls._gateway_client is None:
            cls._gateway_client = GatewayHttpClient.get_instance()
        return cls._gateway_client

    @property
    def gateway_client(self) -> GatewayHttpClient:
        """Instance property to access the gateway client"""
        return self.get_gateway_client()

    def __init__(
        self,
        connector: str,
        trading_pairs: Set[str],
        order_amount_in_base: Decimal,
        update_interval: float = 1.0,
    ) -> None:
        super().__init__()
        self._ev_loop = asyncio.get_event_loop()
        self._price_dict: Dict[str, TokenBuySellPrice] = {}
        self._update_interval = update_interval
        self.fetch_data_loop_task: Optional[asyncio.Task] = None
        # param required for DEX API request
        self.connector = connector
        self.trading_pairs = trading_pairs
        self.order_amount_in_base = order_amount_in_base

        # New format: connector/type (e.g., jupiter/router)
        if "/" not in connector:
            raise ValueError(f"Invalid connector format: {connector}. Use format like 'jupiter/router' or 'uniswap/amm'")
        self._connector_name = connector
        # We'll get chain and network from gateway during price fetching
        self._chain = None
        self._network = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls.dex_logger is None:
            cls.dex_logger = logging.getLogger(__name__)
        return cls.dex_logger

    @property
    def name(self) -> str:
        return f"AmmDataFeed[{self.connector}]"

    @property
    def chain(self) -> str:
        # Chain is determined from gateway
        return self._chain or ""

    @property
    def network(self) -> str:
        # Network is determined from gateway
        return self._network or ""

    @property
    def price_dict(self) -> Dict[str, TokenBuySellPrice]:
        return self._price_dict

    def is_ready(self) -> bool:
        return len(self._price_dict) > 0

    async def check_network(self) -> NetworkStatus:
        is_gateway_online = await self.gateway_client.ping_gateway()
        if not is_gateway_online:
            self.logger().warning("Gateway is not online. Please check your gateway connection.")
        return NetworkStatus.CONNECTED if is_gateway_online else NetworkStatus.NOT_CONNECTED

    async def start_network(self) -> None:
        await self.stop_network()
        self.fetch_data_loop_task = safe_ensure_future(self._fetch_data_loop())

    async def stop_network(self) -> None:
        if self.fetch_data_loop_task is not None:
            self.fetch_data_loop_task.cancel()
            self.fetch_data_loop_task = None

    async def _fetch_data_loop(self) -> None:
        while True:
            try:
                await self._fetch_data()
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(
                    f"Error getting data from {self.name}"
                    f"Check network connection. Error: {e}",
                )
            await self._async_sleep(self._update_interval)

    async def _fetch_data(self) -> None:
        token_price_tasks = [
            asyncio.create_task(self._register_token_buy_sell_price(trading_pair))
            for trading_pair in self.trading_pairs
        ]
        await asyncio.gather(*token_price_tasks, return_exceptions=True)

    async def _register_token_buy_sell_price(self, trading_pair: str) -> None:
        try:
            base, quote = split_hb_trading_pair(trading_pair)
            token_buy_price_task = asyncio.create_task(self._request_token_price(trading_pair, TradeType.BUY))
            token_sell_price_task = asyncio.create_task(self._request_token_price(trading_pair, TradeType.SELL))
            buy_price = await token_buy_price_task
            sell_price = await token_sell_price_task

            if buy_price is not None and sell_price is not None:
                self._price_dict[trading_pair] = TokenBuySellPrice(
                    base=base,
                    quote=quote,
                    connector=self.connector,
                    chain=self._chain or "",
                    network=self._network or "",
                    order_amount_in_base=self.order_amount_in_base,
                    buy_price=buy_price,
                    sell_price=sell_price,
                )
        except Exception as e:
            self.logger().warning(f"Failed to get price for {trading_pair}: {e}")

    async def _request_token_price(self, trading_pair: str, trade_type: TradeType) -> Optional[Decimal]:
        base, quote = split_hb_trading_pair(trading_pair)

        # Use gateway's quote_swap which handles chain/network internally
        try:

            # Get chain and network from connector if not cached
            if not self._chain or not self._network:
                chain, network, error = await self.gateway_client.get_connector_chain_network(
                    self.connector
                )
                if not error:
                    self._chain = chain
                    self._network = network
                else:
                    self.logger().warning(f"Failed to get chain/network for {self.connector}: {error}")
                    return None

            # Use quote_swap which accepts the full connector name
            response = await self.gateway_client.quote_swap(
                network=self._network,
                connector=self.connector,
                base_asset=base,
                quote_asset=quote,
                amount=self.order_amount_in_base,
                side=trade_type,
                slippage_pct=None,
                pool_address=None
            )

            if response and "price" in response:
                return Decimal(str(response["price"]))
            return None
        except Exception as e:
            self.logger().warning(f"Failed to get price using quote_swap: {e}")
            return None

    @staticmethod
    async def _async_sleep(delay: float) -> None:
        """Used to mock in test cases."""
        await asyncio.sleep(delay)



import sys

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType
from hummingbot.core.data_type.in_flight_order import OrderState

CLIENT_ID_PREFIX = "93027a12dac34fBC"
MAX_ID_LEN = 32
SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 30 * 0.8

# URL mapping based on where account is registered:
# sourced from https://app.okx.com/docs-v5/en/#overview-account-mode and https://my.okx.com/docs-v5/en/#overview-account-mode

subdomain_to_api_subdomain = {
    "www": "www",
    "app": "us",
    "my": "eea"
}


def get_okx_base_url(sub_domain: str) -> str:
    """Returns OKX REST base URL based on API subdomain ("www", "us", "eea")"""
    return f"https://{subdomain_to_api_subdomain[sub_domain]}.okx.com/"


def get_ws_url(sub_domain: str) -> str:
    """Returns OKX WebSocket base URL based on API subdomain ("www", "us", "eea")"""
    if sub_domain == "www":
        return "wss://ws.okx.com:8443"
    else:
        return f"wss://ws{subdomain_to_api_subdomain[sub_domain]}.okx.com:8443"


DEFAULT_DOMAIN = get_okx_base_url("www")


def get_okx_ws_uri_public(sub_domain):
    return f"{get_ws_url(sub_domain)}/ws/v5/public"


def get_okx_ws_uri_private(sub_domain):
    return f"{get_ws_url(sub_domain)}/ws/v5/private"


# REST API endpoints
OKX_SERVER_TIME_PATH = '/api/v5/public/time'
OKX_INSTRUMENTS_PATH = '/api/v5/public/instruments'
OKX_TICKER_PATH = '/api/v5/market/ticker'
OKX_TICKERS_PATH = '/api/v5/market/tickers'
OKX_ORDER_BOOK_PATH = '/api/v5/market/books'

# Auth required
OKX_PLACE_ORDER_PATH = "/api/v5/trade/order"
OKX_ORDER_DETAILS_PATH = '/api/v5/trade/order'
OKX_ORDER_CANCEL_PATH = '/api/v5/trade/cancel-order'
OKX_BATCH_ORDER_CANCEL_PATH = '/api/v5/trade/cancel-batch-orders'
OKX_BALANCE_PATH = '/api/v5/account/balance'
OKX_TRADE_FILLS_PATH = "/api/v5/trade/fills"

# WebSocket channels
OKX_WS_ACCOUNT_CHANNEL = "account"
OKX_WS_ORDERS_CHANNEL = "orders"
OKX_WS_PUBLIC_TRADES_CHANNEL = "trades"
OKX_WS_PUBLIC_BOOKS_CHANNEL = "books"

OKX_WS_CHANNELS = {
    OKX_WS_ACCOUNT_CHANNEL,
    OKX_WS_ORDERS_CHANNEL
}

# Rate limiting
WS_CONNECTION_LIMIT_ID = "WSConnection"
WS_REQUEST_LIMIT_ID = "WSRequest"
WS_SUBSCRIPTION_LIMIT_ID = "WSSubscription"
WS_LOGIN_LIMIT_ID = "WSLogin"

ORDER_STATE = {
    "live": OrderState.OPEN,
    "filled": OrderState.FILLED,
    "partially_filled": OrderState.PARTIALLY_FILLED,
    "canceled": OrderState.CANCELED,
}

ORDER_TYPE_MAP = {
    OrderType.LIMIT: "limit",
    OrderType.MARKET: "market",
    OrderType.LIMIT_MAKER: "post_only",
}

NO_LIMIT = sys.maxsize

RATE_LIMITS = [
    RateLimit(WS_CONNECTION_LIMIT_ID, limit=3, time_interval=1),
    RateLimit(WS_REQUEST_LIMIT_ID, limit=100, time_interval=10),
    RateLimit(WS_SUBSCRIPTION_LIMIT_ID, limit=240, time_interval=60 * 60),
    RateLimit(WS_LOGIN_LIMIT_ID, limit=1, time_interval=15),
    RateLimit(limit_id=OKX_SERVER_TIME_PATH, limit=10, time_interval=2),
    RateLimit(limit_id=OKX_INSTRUMENTS_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_TICKER_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_TICKERS_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_ORDER_BOOK_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_PLACE_ORDER_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_ORDER_DETAILS_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_ORDER_CANCEL_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_BATCH_ORDER_CANCEL_PATH, limit=300, time_interval=2),
    RateLimit(limit_id=OKX_BALANCE_PATH, limit=10, time_interval=2),
    RateLimit(limit_id=OKX_TRADE_FILLS_PATH, limit=60, time_interval=2),
]



from decimal import Decimal
from typing import Optional

from hummingbot.connector.utils import split_hb_trading_pair

s_decimal_0 = Decimal(0)
s_decimal_max = Decimal("1e56")
s_decimal_min = Decimal(1) / s_decimal_max


cdef class TradingRule:
    def __init__(self,
                 trading_pair: str,
                 min_order_size: Decimal = s_decimal_0,
                 max_order_size: Decimal = s_decimal_max,
                 min_price_increment: Decimal = s_decimal_min,
                 min_base_amount_increment: Decimal = s_decimal_min,
                 min_quote_amount_increment: Decimal = s_decimal_min,
                 min_notional_size: Decimal = s_decimal_0,
                 min_order_value: Decimal = s_decimal_0,
                 max_price_significant_digits: Decimal = s_decimal_max,
                 supports_limit_orders: bool = True,
                 supports_market_orders: bool = True,
                 buy_order_collateral_token: Optional[str] = None,
                 sell_order_collateral_token: Optional[str] = None):
        self.trading_pair = trading_pair
        self.min_order_size = min_order_size
        self.max_order_size = max_order_size
        self.min_price_increment = min_price_increment
        self.min_base_amount_increment = min_base_amount_increment
        self.min_quote_amount_increment = min_quote_amount_increment
        self.min_notional_size = min_notional_size
        self.min_order_value = min_order_value
        self.max_price_significant_digits = max_price_significant_digits
        self.supports_limit_orders = supports_limit_orders
        self.supports_market_orders = supports_market_orders
        quote_token = split_hb_trading_pair(self.trading_pair)[1]
        self.buy_order_collateral_token = buy_order_collateral_token or quote_token
        self.sell_order_collateral_token = sell_order_collateral_token or quote_token

    def __repr__(self) -> str:
        return f"TradingRule(trading_pair='{self.trading_pair}', " \
               f"min_order_size={self.min_order_size}, " \
               f"max_order_size={self.max_order_size}, " \
               f"min_price_increment={self.min_price_increment}, " \
               f"min_base_amount_increment={self.min_base_amount_increment}, " \
               f"min_quote_amount_increment={self.min_quote_amount_increment}, " \
               f"min_notional_size={self.min_notional_size}, " \
               f"min_order_value={self.min_order_value}, " \
               f"max_price_significant_digits={self.max_price_significant_digits}, " \
               f"supports_limit_orders={self.supports_limit_orders}, " \
               f"supports_market_orders={self.supports_market_orders}, " \
               f"buy_order_collateral_token={self.buy_order_collateral_token}, " \
               f"sell_order_collateral_token={self.sell_order_collateral_token}," \
               f")"
